# Copyright 2013 Christoph Reiter
#           2020 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet import app
from quodlibet import config
from quodlibet.qltk.models import ObjectStore, ObjectModelFilter
from quodlibet.qltk.models import ObjectModelSort


class AlbumItem:
    cover = None
    scanned = False

    def __init__(self, album):
        self.album = album

    @property
    def cover_size(self):
        size = config.getint("browsers", "cover_size")
        if size <= 0:
            size = 48
        return size

    def scan_cover(self, force=False, scale_factor=1, callback=None, cancel=None):
        if (self.scanned and not force) or not self.album or not self.album.songs:
            return
        self.scanned = True

        def set_cover_cb(pixbuf):
            self.cover = pixbuf
            callback()

        s = self.cover_size * scale_factor
        app.cover_manager.get_pixbuf_many_async(
            self.album.songs, s, s, cancel, set_cover_cb
        )

    def __repr__(self):
        return repr(self.album)


class AlbumModelMixin:
    def get_items(self, paths):
        items = []
        for path in paths:
            item = self.get_value(self.get_iter(path))
            if item.album is None:
                return [i for i in self.itervalues() if i.album is not None]
            items.append(item)
        return items

    def get_albums(self, paths):
        return [i.album for i in self.get_items(paths)]

    def get_album(self, iter_):
        return self.get_value(iter_, 0).album


class AlbumModel(ObjectStore, AlbumModelMixin):
    def __init__(self, library):
        super().__init__()
        self.__library = library

        albums = library.albums
        self.__sigs = [
            albums.connect("added", self._add_albums),
            albums.connect("removed", self._remove_albums),
            albums.connect("changed", self._change_albums),
        ]

        self.append(row=[AlbumItem(None)])
        self.append_many(AlbumItem(a) for a in albums.values())

    def refresh_all(self):
        """Trigger redraws for all rows"""

        for iter_, _value in self.iterrows():
            self.row_changed(self.get_path(iter_), iter_)

    def destroy(self):
        library = self.__library
        for sig in self.__sigs:
            library.albums.disconnect(sig)
        self.__library = None
        self.clear()

    def _update_all(self):
        if not self.is_empty():
            row = self[0]
            self.row_changed(row.path, row.iter)

    def _add_albums(self, library, added):
        self.append_many(AlbumItem(a) for a in added)
        self._update_all()

    def _remove_albums(self, library, removed):
        removed_albums = removed.copy()
        iters_remove = []
        for iter_, item in self.iterrows():
            album = item.album
            if album is not None and album in removed_albums:
                removed_albums.remove(album)
                iters_remove.append(iter_)
                if not removed_albums:
                    break

        for iter_ in iters_remove:
            self.remove(iter_)

        self._update_all()

    def _change_albums(self, library, changed):
        """Trigger a row redraw for each album that changed"""

        changed_albums = changed.copy()
        for iter_, item in self.iterrows():
            album = item.album
            if album is not None and album in changed_albums:
                changed_albums.remove(album)
                self.row_changed(self.get_path(iter_), iter_)
                if not changed_albums:
                    break


class AlbumFilterModel(ObjectModelFilter, AlbumModelMixin):
    def contains_all(self, paths):
        values = (self.get_value(self.get_iter(p), 0).album for p in paths)
        return None in values


class AlbumSortModel(ObjectModelSort, AlbumModelMixin):
    pass



# Copyright 2018-2023 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import operator
import os
import shutil
from functools import reduce
from collections.abc import Iterable

from gi.repository import GObject, Gtk, Gdk, Gio, GLib, Soup, GdkPixbuf

from quodlibet import _, app, print_d, print_w, util
from quodlibet import qltk
from quodlibet.formats import AudioFile
from senf import path2fsn
from quodlibet.pattern import ArbitraryExtensionFileFromPattern, Pattern
from quodlibet.plugins import PluginConfig, ConfProp, IntConfProp, BoolConfProp
from quodlibet.plugins.songshelpers import any_song, is_a_file
from quodlibet.plugins.songsmenu import SongsMenuPlugin
from quodlibet.qltk import Icons
from quodlibet.qltk.paned import Paned
from quodlibet.qltk.window import PersistentWindowMixin
from quodlibet.util import connect_destroy, format_size, escape
from quodlibet.util.cover import CoverManager
from quodlibet.util.cover.manager import CoverData
from quodlibet.util.http import download

_FULL_PAT = "<albumartist|<albumartist>|<artist>><album| - <album>| - <title>>"
SAVE_PATTERNS = [
    "folder",
    "cover",
    _FULL_PAT,
    f"{_FULL_PAT} - front",
    f"{_FULL_PAT} - back",
]

IMAGE_EXTENSIONS = {
    "image/jpeg": "jpg",
    "image/png": "png",
    "image/gif": "gif",
}


class DownloadCoverArt(SongsMenuPlugin):
    """Download and save album (cover) art from a variety of sources"""

    PLUGIN_ID = "Download Cover Art"
    PLUGIN_NAME = _("Download Cover Art")
    PLUGIN_DESC = _(
        "Downloads high-quality album covers using Quod Libet cover plugins."
    )
    PLUGIN_ICON = Icons.INSERT_IMAGE
    REQUIRES_ACTION = True

    plugin_handles = any_song(lambda song: song("album") and is_a_file(song))

    def plugin_album(self, songs):
        manager = app.cover_manager
        if not (songs and songs[0]("album")):
            return
        dialog = CoverArtWindow(songs, manager, Config())
        ret = dialog.run()
        if ret == Gtk.ResponseType.APPLY:
            manager.cover_changed(songs)
        dialog.destroy()


class Config:
    plugin_config = PluginConfig(DownloadCoverArt.PLUGIN_ID.replace(" ", "_"))
    preview_size = IntConfProp(plugin_config, "preview_size", 500)
    save_pattern = ConfProp(plugin_config, "save_pattern", "folder.%(ext)")
    over_scale = BoolConfProp(plugin_config, "preview_over_scale", True)
    re_encode = BoolConfProp(plugin_config, "re_encode", False)
    jpg_quality = IntConfProp(plugin_config, "jpeg_quality", 95)


class ResizeWebImage(Gtk.Image):
    """A resizeable widget that can receive loaded images, and save them"""

    __gsignals__ = {
        # The content-type, size (in bytes) and properties of a cover
        # once discovered
        "info-known": (GObject.SignalFlags.RUN_LAST, None, (str, int, object)),
        "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
    }

    def __init__(self, url, config: Config, cancellable=None):
        super().__init__()
        self.config = config
        self.url = url
        self.cancellable = cancellable
        self.message = msg = Soup.Message.new("GET", self.url)
        self._content_type = None
        self._original = None
        download(
            msg,
            cancellable,
            self._sent,
            None,
            failure_callback=lambda *args: self.emit("failed", self.url),
        )
        self.set_size_request(config.preview_size, config.preview_size)
        self._pixbuf = None

    @property
    def extension(self):
        return (
            "jpg"
            if self.config.re_encode
            else IMAGE_EXTENSIONS.get(self._content_type, "jpg")
        )

    def _sent(self, msg, result, data):
        headers = self.message.get_response_headers()
        self.size = int(headers.get_content_length())
        self._content_type = headers.get_content_type()[0]
        self._original = result
        try:
            loader = GdkPixbuf.PixbufLoader()
        except GLib.GError as e:
            print_w(f"Couldn't create GdkPixbuf ({e})")
        else:
            loader.write(result)
            loader.close()
            self._pixbuf = loader.get_pixbuf()
            self.emit("info-known", self._content_type, self.size, self._pixbuf.props)
            self.resize()
            self.queue_draw()

    def resize(self, new_size=None):
        if not self._pixbuf:
            return
        if new_size:
            self.config.preview_size = new_size
        else:
            new_size = self.config.preview_size
        props = self._pixbuf.props
        if not self.config.over_scale:
            new_size = min(props.width, props.height, new_size)
        resized = self._pixbuf.scale_simple(
            new_size, new_size, GdkPixbuf.InterpType.BILINEAR
        )
        self.set_from_pixbuf(resized)
        self.set_size_request(new_size, new_size)

    def save_image(self, fsn):
        fsn = path2fsn(fsn)
        if self.config.re_encode:
            quality = str(self.config.jpg_quality)
            print_d(f"Converting image to JPEG @ {quality}%")
            ret = self._pixbuf.savev(fsn, "jpeg", ["quality"], [quality])
            if not ret:
                raise OSError(f"Couldn't save to {fsn}")
        else:
            print_d(f"Saving original image to {fsn}")
            with open(fsn, "wb") as f:
                f.write(self._original)


class CoverArtWindow(qltk.Dialog, PersistentWindowMixin):
    SIZES = {
        300: _("Small"),
        500: _("Classic"),
        600: _("Large"),
        720: _("HD"),
        1080: _("Full HD"),
        1600: _("WQXGA"),
        2160: _("4K UHD"),
    }
    DEFAULT_SIZE = list(SIZES.keys())[0]

    def __init__(
        self,
        songs: Iterable[AudioFile],
        manager: CoverManager,
        config: Config,
        headless: bool = False,
        **kwargs,
    ):
        super().__init__(title=_("Cover Art Download"), use_header_bar=True, **kwargs)
        self.set_default_size(1400, 720)
        id_ = DownloadCoverArt.PLUGIN_ID.lower().replace(" ", "_")
        self.enable_window_tracking(id_)
        self.config = config
        self.headless = headless
        self.flow_box = box = Gtk.FlowBox()
        self.model = Gio.ListStore()
        self.songs = songs
        self._groups = {}
        self.button = None

        def selected(fb):
            children = fb.get_selected_children()
            if children:
                cover = self.model.get_item(children[0].get_index())
                if cover:
                    self.button.set_sensitive(True)

        box.set_valign(Gtk.Align.START)
        box.set_max_children_per_line(4)
        box.connect("selected-children-changed", selected)

        paned = Paned(orientation=Gtk.Orientation.VERTICAL)
        paned.ensure_wide_handle()
        sw = Gtk.ScrolledWindow()
        sw.add(self.flow_box)

        paned.pack1(sw, True, True)
        paned.pack2(self.create_options(), False, False)
        self.vbox.pack_start(paned, True, True, 0)

        connect_destroy(manager, "covers-found", self._covers_found)
        connect_destroy(manager, "searches-complete", self._finished)
        cancellable = self.__cancellable = Gio.Cancellable()
        self.show_all()

        # Do the search
        self._groups = manager.search_cover(cancellable, songs)

    def destroy(self):
        if self.__cancellable:
            print_d("Cancelling remaining requests...")
            self.__cancellable.cancel()
        super().destroy()

    def _image_failed(self, _view, message: str, widget: Gtk.Widget):
        print_d(f"Failed downloading image ({message}), removing result.")
        # FlowBox creates a hidden child on addition
        self.flow_box.remove(widget.get_parent())

    def _create_item_widget(self, item):
        def update(img, content_type, size, props, item, frame):
            format = IMAGE_EXTENSIONS.get(content_type, content_type).upper()
            source = escape(item.source)
            text = (
                f"{source} - {format}, "
                f"{props.width} x {props.height}, "
                f"{util.bold(format_size(size))}"
            )
            frame.get_label_widget().set_markup(text)
            frame.get_child().set_reveal_child(True)

        img = ResizeWebImage(item.url, config=self.config)
        text = _("Loading %(source)s - %(dimensions)s…") % {
            "source": item.source,
            "dimensions": item.dimensions,
        }
        frame = Gtk.Frame.new(text)
        img.set_padding(12, 12)
        frame.set_shadow_type(Gtk.ShadowType.NONE)
        frame.set_border_width(12)
        img.connect("info-known", update, item, frame)
        img.connect("failed", self._image_failed, frame)
        reveal = Gtk.Revealer()
        reveal.set_reveal_child(False)
        reveal.props.transition_duration = 800
        reveal.props.transition_type = Gtk.RevealerTransitionType.CROSSFADE

        eb = Gtk.EventBox()
        eb.add(img)
        reveal.add(eb)
        frame.add(reveal)
        frame.set_label_align(0.5, 1.0)
        reveal.connect("button-press-event", self._on_click)
        return frame

    def _on_click(self, view, event):
        # TODO: less hacky way to detect double-click
        if (
            event.button == Gdk.BUTTON_PRIMARY
            and event.type != Gdk.EventType.BUTTON_PRESS
        ):
            self.__save(None)
            self.destroy()

    def _filenames(self, pat_text, ext, full_path=False) -> list[str]:
        def fn_for(song):
            pat = ArbitraryExtensionFileFromPattern(f"{pat_text}.{ext}")
            fn = pat.format(song)
            return os.path.join(song("~dirname"), fn) if full_path else fn

        return sorted({fn_for(song) for song in self.songs})

    def _covers_found(self, manager, provider, results):
        if not results:
            print_d(f"No results from {provider}")
            return
        for result in results:
            self.model.append(result)
            self.flow_box.insert(self._create_item_widget(result), 0)
        self.show_all()

    def _finished(self, manager, results):
        if not any(results.values()):
            print_w(f"Nothing found from {len(self._groups)} provider(s)")

            self.button.set_sensitive(False)
            if not self.headless:
                self._quit(results)

    def _quit(self, results):
        pat = Pattern("<albumartist|<albumartist>|<artist>> - <album>")
        group_songs = [
            songs for group in self._groups.values() for songs in group.values()
        ]
        texts = {pat.format(s) for s in reduce(operator.concat, group_songs, [])}
        albums = "\n".join(texts)
        providers = ", ".join({manager.name for manager in results.keys()})
        data = {"albums": util.italic(albums), "providers": util.monospace(providers)}
        markup = (
            _(
                "Nothing found for albums:\n%(albums)s.\n\n"
                "Providers used:\n%(providers)s"
            )
            % data
        )
        dialog = qltk.Message(
            Gtk.MessageType.INFO,
            parent=self,
            title=_("No covers found"),
            description=markup,
            escape_desc=False,
        )
        dialog.run()
        self.destroy()

    @staticmethod
    def __image_from_child(child):
        # Ugh, horrible
        return child.get_child().get_child().get_child().get_child()

    def create_options(self):
        frame = Gtk.Frame(label=_("Options"))
        hbox = Gtk.HBox()
        sizes = self.SIZES.keys()
        slider = Gtk.Scale.new_with_range(
            Gtk.Orientation.HORIZONTAL, min(sizes), max(sizes), 100
        )
        for size, name in self.SIZES.items():
            slider.add_mark(size, Gtk.PositionType.BOTTOM, name)
        slider.set_show_fill_level(False)
        slider.set_value(self.config.preview_size)

        def format_dims(_slider, value):
            size = int(value)
            return _("%(size)d ✕ %(size)d px") % {"size": size}

        def slider_changed(_slider):
            new_size = slider.get_value()
            try:
                for child in self.flow_box.get_children():
                    img = self.__image_from_child(child)
                    img.resize(new_size)
            except AttributeError as e:
                print_w(f"Couldn't set picture size(s) ({e})")

        slider.connect("format-value", format_dims)
        slider.connect("value-changed", slider_changed)
        label = Gtk.Label(_("Preview size"))
        label.set_mnemonic_widget(slider)
        hbox.pack_start(label, False, False, 6)
        hbox.pack_start(slider, True, True, 6)
        vbox = Gtk.VBox()
        vbox.pack_start(hbox, False, False, 6)

        def create_save_box():
            hbox = Gtk.HBox()
            label = Gtk.Label(_("Save destination"))
            hbox.pack_start(label, False, False, 6)
            model = Gtk.ListStore(str)
            for val in SAVE_PATTERNS:
                model.append(row=[val])
            save_filename = Gtk.ComboBox(model=model)
            label.set_mnemonic_widget(save_filename)
            cell = Gtk.CellRendererText()
            save_filename.pack_start(cell, True)

            def draw_save_type(column, cell, model, it, data):
                pat_text = model[it][0]
                ext = "jpg" if self.config.re_encode else "*"
                text = list(self._filenames(pat_text, ext))[0]
                cell.set_property("markup", util.monospace(text))

            save_filename.set_cell_data_func(cell, draw_save_type, None)

            def changed(combo):
                value = model.get_value(combo.get_active_iter(), 0)
                self.config.save_pattern = value

            def select_value(combo, value):
                for i, item in enumerate(model):
                    if value == item[0]:
                        combo.set_active(i)

            save_filename.connect("changed", changed)
            select_value(save_filename, self.config.save_pattern)
            hbox.pack_start(save_filename, False, False, 6)
            create_ccb = self.config.plugin_config.ConfigCheckButton
            tooltip = _(
                "If not already a JPEG, convert the image to "
                "a high-quality JPEG of the same size"
            )
            re_encode = create_ccb(
                _("Save as JPEG"), "re_encode", tooltip=tooltip, populate=True
            )
            re_encode.connect("toggled", lambda _: hbox.queue_draw())
            hbox.pack_start(re_encode, False, False, 6)
            return hbox

        vbox.pack_start(create_save_box(), False, False, 6)
        frame.add(vbox)

        self.button = self.add_icon_button(
            _("_Save"), Icons.DOCUMENT_SAVE, Gtk.ResponseType.APPLY
        )
        self.button.set_sensitive(False)
        self.button.connect("clicked", self.__save)
        return frame

    def __save(self, button):
        child = self.flow_box.get_selected_children()[0]
        data = self.model.get_item(child.get_index())
        img = self.__image_from_child(child)
        self._save_images(data, img)

    def _save_images(self, data: CoverData, img: Gtk.Image):
        paths = self._filenames(self.config.save_pattern, img.extension, full_path=True)
        try:
            first_path = paths.pop()
        except IndexError:
            print_w("No paths to save somewhow", self)
        print_d(f"Saving {data} to {first_path}")
        img.save_image(first_path)
        # Copying faster than potentially resizing
        if paths:
            for path in paths:
                shutil.copy(first_path, path)
            print_d(f"Finished copying to {paths}")



# Copyright 2004-2005 Joe Wreschnig, Michael Urman, Iñigo Serna
#           2016-2023 Nick Boultbee
#                2017 Fredrik Strupe
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import time
from typing import Any
from collections.abc import Sequence

from gi.repository import Gtk, Gdk, Pango, GLib

from quodlibet.library.base import Library
from quodlibet.player._base import BasePlayer
from senf import bytes2fsn, fsn2bytes

import quodlibet
from quodlibet import ngettext, _, print_e, print_w, print_d
from quodlibet import config
from quodlibet import qltk
from quodlibet import app

from quodlibet.util import (
    connect_destroy,
    connect_after_destroy,
    format_time_preferred,
    print_exc,
    DeferredSignal,
)
from quodlibet.qltk import Icons, gtk_version, add_css
from quodlibet.qltk.ccb import ConfigCheckMenuItem
from quodlibet.qltk.songlist import SongList, DND_QL, DND_URI_LIST
from quodlibet.qltk.songsmenu import SongsMenu
from quodlibet.qltk.menubutton import SmallMenuButton
from quodlibet.qltk.songmodel import PlaylistModel
from quodlibet.qltk.playorder import OrderInOrder, OrderShuffle
from quodlibet.qltk.x import (
    ScrolledWindow,
    SymbolicIconImage,
    SmallImageButton,
    SmallImageToggleButton,
    MenuItem,
    RadioMenuItem,
)
from quodlibet.qltk.songlistcolumns import CurrentColumn

QUEUE = os.path.join(quodlibet.get_user_dir(), "queue")


class PlaybackStatusIcon(Gtk.Box):
    """A widget showing a play/pause/stop symbolic icon"""

    def __init__(self):
        super().__init__()
        self._icons = {}

    def _set(self, name):
        if name not in self._icons:
            image = SymbolicIconImage(name, Gtk.IconSize.MENU)
            self._icons[name] = image
            image.show()
        else:
            image = self._icons[name]

        children = self.get_children()
        if children:
            self.remove(children[0])
        self.add(image)

    def play(self):
        self._set("media-playback-start")

    def stop(self):
        self._set("media-playback-stop")

    def pause(self):
        self._set("media-playback-pause")


class ExpandBoxHack(Gtk.HBox):
    def do_get_preferred_width(self):
        # Workaround for https://bugzilla.gnome.org/show_bug.cgi?id=765602
        # set_label_fill() no longer works since 3.20. Fake a natural size
        # which is larger than the expander can be to force the parent to
        # allocate to us the whole space.
        min_, nat = Gtk.HBox.do_get_preferred_width(self)
        if gtk_version > (3, 19):
            # if we get too large gtk calcs will overflow..
            nat = max(nat, 2**16)
        return (min_, nat)


class QueueExpander(Gtk.Expander):
    def __init__(self, library, player):
        super().__init__(spacing=3)
        sw = ScrolledWindow()
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        sw.set_shadow_type(Gtk.ShadowType.IN)
        save_interval_secs = config.getint("autosave", "queue_interval")
        self.queue = PlayQueue(library, player, save_interval_secs)
        self.queue.props.expand = True
        sw.add(self.queue)

        add_css(self, ".ql-expanded title { margin-bottom: 5px; }")

        outer = ExpandBoxHack()

        left = Gtk.HBox(spacing=12)

        hb2 = Gtk.HBox(spacing=3)
        state_icon = PlaybackStatusIcon()
        state_icon.stop()
        state_icon.show()
        hb2.pack_start(state_icon, True, True, 0)
        name_label = Gtk.Label(label=_("_Queue"), use_underline=True)
        name_label.set_size_request(-1, 24)
        hb2.pack_start(name_label, True, True, 0)
        left.pack_start(hb2, False, True, 0)

        menu = Gtk.Menu()

        self.count_label = count_label = Gtk.Label()
        self.count_label.set_property("ellipsize", Pango.EllipsizeMode.END)
        self.count_label.set_width_chars(10)
        self.count_label.get_style_context().add_class("dim-label")
        left.pack_start(count_label, False, True, 0)

        outer.pack_start(left, True, True, 0)

        self.set_label_fill(True)

        clear_item = SmallImageButton(
            image=SymbolicIconImage(Icons.USER_TRASH, Gtk.IconSize.MENU),
            relief=Gtk.ReliefStyle.NONE,
            tooltip_text=_("Clear Queue"),
        )
        clear_item.connect("clicked", self.__clear_queue)
        outer.pack_start(clear_item, False, False, 3)

        toggle = SmallImageToggleButton(
            image=SymbolicIconImage(Icons.SYSTEM_LOCK_SCREEN, Gtk.IconSize.MENU),
            relief=Gtk.ReliefStyle.NONE,
            tooltip_text=_("Disable queue - the queue will be ignored when playing"),
        )
        disabled = config.getboolean("memory", "queue_disable", False)
        toggle.props.active = disabled
        self.__queue_disable(disabled)
        toggle.connect("toggled", lambda b: self.__queue_disable(b.props.active))
        outer.pack_start(toggle, False, False, 3)

        mode_menu = Gtk.Menu()

        norm_mode_item = RadioMenuItem(
            label=_("Ephemeral"),
            tooltip_text=_("Remove songs from the queue after playing them"),
            group=None,
        )
        mode_menu.append(norm_mode_item)
        norm_mode_item.set_active(True)
        norm_mode_item.connect("toggled", lambda _: self.__keep_songs_enable(False))

        keep_mode_item = RadioMenuItem(
            label=_("Persistent"),
            tooltip_text=_("Keep songs in the queue after playing them"),
            group=norm_mode_item,
        )
        mode_menu.append(keep_mode_item)
        keep_mode_item.connect("toggled", lambda b: self.__keep_songs_enable(True))
        keep_mode_item.set_active(
            config.getboolean("memory", "queue_keep_songs", False)
        )

        mode_item = MenuItem(_("Mode"), Icons.SYSTEM_RUN)
        mode_item.set_submenu(mode_menu)
        menu.append(mode_item)

        rand_checkbox = ConfigCheckMenuItem(
            _("_Random"), "memory", "shufflequeue", populate=True
        )
        rand_checkbox.connect("toggled", self.__queue_shuffle)
        self.set_shuffled(rand_checkbox.get_active())
        menu.append(rand_checkbox)

        stop_checkbox = ConfigCheckMenuItem(
            _("Stop at End"), "memory", "queue_stop_at_end", populate=True
        )
        menu.append(stop_checkbox)

        button = SmallMenuButton(
            SymbolicIconImage(Icons.EMBLEM_SYSTEM, Gtk.IconSize.MENU), arrow=True
        )
        button.set_relief(Gtk.ReliefStyle.NORMAL)
        button.show_all()
        button.hide()
        button.set_no_show_all(True)
        menu.show_all()
        button.set_menu(menu)

        outer.pack_start(button, False, False, 3)

        close_button = SmallImageButton(
            image=SymbolicIconImage("window-close", Gtk.IconSize.MENU),
            relief=Gtk.ReliefStyle.NONE,
        )

        close_button.connect("clicked", lambda *x: self.hide())

        outer.pack_start(close_button, False, False, 6)

        self.set_label_widget(outer)
        self.add(sw)
        self.connect("notify::expanded", self.__expand, button)
        self.connect("notify::expanded", self.__expand, button)

        targets = [
            ("text/x-quodlibet-songs", Gtk.TargetFlags.SAME_APP, DND_QL),
            ("text/uri-list", 0, DND_URI_LIST),
        ]
        targets = [Gtk.TargetEntry.new(*t) for t in targets]

        self.drag_dest_set(Gtk.DestDefaults.ALL, targets, Gdk.DragAction.COPY)
        self.connect("drag-motion", self.__motion)
        self.connect("drag-data-received", self.__drag_data_received)

        self.queue.model.connect_after(
            "row-inserted", DeferredSignal(self.__check_expand), count_label
        )
        self.queue.model.connect_after(
            "row-deleted", DeferredSignal(self.__update_count), count_label
        )

        self.__update_count(self.model, None, count_label)

        connect_destroy(player, "song-started", self.__update_state_icon, state_icon)
        connect_destroy(
            player, "paused", self.__update_state_icon_pause, state_icon, True
        )
        connect_destroy(
            player, "unpaused", self.__update_state_icon_pause, state_icon, False
        )

        connect_destroy(player, "song-started", self.__song_started, self.queue.model)
        connect_destroy(
            player, "song-ended", self.__update_queue_stop, self.queue.model
        )

        # to make the children clickable if mapped
        # ....no idea why, but works
        def hack(expander):
            label = expander.get_label_widget()
            if label:
                label.unmap()
                label.map()

        self.connect("map", hack)

        self.set_expanded(config.getboolean("memory", "queue_expanded"))
        self.notify("expanded")

        for child in self.get_children():
            child.show_all()

    @property
    def model(self):
        return self.queue.model

    def refresh(self):
        self.__update_count(self.model, None, self.count_label)

    def __update_state_icon(self, player, song, state_icon):
        if self.model.sourced:
            state_icon.play()
        else:
            state_icon.stop()

    def __update_state_icon_pause(self, player, state_icon, paused):
        if self.model.sourced:
            if paused:
                state_icon.pause()
            else:
                state_icon.play()
        else:
            state_icon.stop()

    def __clear_queue(self, activator):
        self.model.clear()

    def __keep_songs_enable(self, enabled):
        config.set("memory", "queue_keep_songs", enabled)
        if enabled:
            self.queue.set_first_column_type(CurrentColumn)
        else:
            for col in self.queue.get_columns():
                # Remove the CurrentColum if it exists
                if isinstance(col, CurrentColumn):
                    self.queue.set_first_column_type(None)
                    break

    def __motion(self, wid, context, x, y, time):
        Gdk.drag_status(context, Gdk.DragAction.COPY, time)
        return True

    def __update_count(self, model, path, lab):
        if len(model) == 0:
            text = ""
        else:
            time = sum([row[0].get("~#length", 0) for row in model])
            text = ngettext(
                "%(count)d song (%(time)s)", "%(count)d songs (%(time)s)", len(model)
            ) % {"count": len(model), "time": format_time_preferred(time)}
        lab.set_text(text)

    def __check_expand(self, model, path, iter, lab):
        self.__update_count(model, path, lab)
        self.show()

    def __drag_data_received(self, expander, *args):
        self.queue.emit("drag-data-received", *args)

    def __queue_shuffle(self, button):
        self.set_shuffled(button.get_active())

    def set_shuffled(self, is_shuffled):
        self.queue.model.order = OrderShuffle() if is_shuffled else OrderInOrder()

    def __update_queue_stop(self, player, song, stopped, model):
        enabled = config.getboolean("memory", "queue_stop_at_end", False)
        songs_left = len(model.get())
        if enabled and not stopped and songs_left == 0 and self._queue_sourced:
            app.player.stop()

    def __song_started(self, player, song, model):
        self._queue_sourced = model.sourced

    def __expand(self, widget, prop, menu_button):
        expanded = self.get_expanded()

        style_context = self.get_style_context()
        if expanded:
            style_context.add_class("ql-expanded")
        else:
            style_context.remove_class("ql-expanded")

        menu_button.set_property("visible", expanded)
        config.set("memory", "queue_expanded", str(expanded))

    def __queue_disable(self, disabled):
        config.set("memory", "queue_disable", disabled)
        self.queue.set_sensitive(not disabled)


class QueueModel(PlaylistModel):
    """
    A play list model for queues.

    In contrast to `PlaylistModel`, when new songs have been set, its play order
    will disregard the current song from the old set of songs and start from
    scratch.
    """

    __reset = False

    def set(self, songs: Sequence[Any]):
        self.__reset = True
        super().set(songs)

    def next(self):
        print_d(f"Using {self.order}.next_explicit() to get next song")

        iter_ = None if self.__reset else self.current_iter
        self.__reset = False
        self.current_iter = self.order.next_explicit(self, iter_)

    def previous(self):
        iter_ = None if self.__reset else self.current_iter
        self.__reset = False
        self.current_iter = self.order.previous_explicit(self, iter_)

    def go_to(self, song_or_iter, explicit=False, source=None):
        self.__reset = False
        return super().go_to(song_or_iter, explicit, source)


class PlayQueue(SongList):
    _activated = False
    _MAX_PENDING = 5
    """Maximum number of queue items to leave unpersisted (batching)"""

    def __init__(
        self,
        library: Library,
        player: BasePlayer,
        autosave_interval_secs: int | None = None,
    ):
        super().__init__(library, player, model_cls=QueueModel, sortable=False)
        self._updated_time = time.time()
        self.autosave_interval = autosave_interval_secs
        self._pending = 0
        keep_song = config.getboolean("memory", "queue_keep_songs", False)
        if keep_song:
            self.set_first_column_type(CurrentColumn)
        self.set_size_request(-1, 120)
        self.connect("row-activated", self.__go_to, player)

        def reset_activated(*args):
            self._activated = False

        connect_after_destroy(player, "song-started", reset_activated)

        self.connect("popup-menu", self.__popup, library)
        self.enable_drop()

        def write(*args, **kwargs):
            self._write(self, self.model)
            return True

        self.connect("destroy", self.__destroy)
        if self.autosave_interval:
            self._tid = GLib.timeout_add(self.autosave_interval * 1000, write)
        else:
            self._tid = None
        connect_after_destroy(self.model, "row-inserted", write)
        connect_after_destroy(self.model, "row-deleted", write)
        self.__fill(library)

        self.connect("key-press-event", self.__delete_key_pressed)

    def __destroy(self, widget):
        self._write(widget, self.model, force=True)
        if self._tid:
            GLib.source_remove(self._tid)
            self._tid = None
            print_d("Stopped autosave")

    def __delete_key_pressed(self, widget, event):
        if qltk.is_accel(event, "Delete"):
            self.__remove()
            return True
        return False

    def __go_to(self, view, path, column, player):
        self._activated = True
        if player.go_to(self.model.get_iter(path), explicit=True, source=self.model):
            player.paused = False

    def __fill(self, library):
        try:
            with open(QUEUE, "rb") as f:
                lines = f.readlines()
        except OSError:
            return

        filenames = []
        for line in lines:
            line = line.strip()
            if not line:
                continue
            try:
                filename = bytes2fsn(line, "utf-8")
            except ValueError:
                print_exc()
                continue
            filenames.append(filename)

        if library.librarian:
            library = library.librarian
        songs = filter(None, map(library.get, filenames))
        for song in songs:
            self.model.append([song])

    def _write(self, _widget: Gtk.Widget, model: QueueModel, force=False):
        diff = time.time() - self._updated_time
        if not self._should_write(force, diff):
            self._pending += 1
            return
        print_d(f"Saving play queue after {diff:.1f}s")
        filenames = [row[0]["~filename"] for row in model]
        try:
            with open(QUEUE, "wb") as f:
                for filename in filenames:
                    try:
                        line = fsn2bytes(filename, "utf-8")
                    except ValueError as e:
                        print_w(f"Ignoring queue save error ({e})")
                        continue
                    f.write(line + b"\n")
        except OSError as e:
            print_e(f"Error saving queue ({e})")
        self._updated_time = time.time()
        self._pending = 0

    def _should_write(self, force: bool, diff: float):
        """Whether it's appropriate to write (flush)
        Either it's disabled by config, in which case we use batching,
        or we respect the last write time and the delta
        """
        if force:
            return True
        if self.autosave_interval:
            # Generally only save if there are *known* pending updates,
            # but these don´t catch everything, so save regardless every now and again.
            return (
                diff > self.autosave_interval and self._pending
            ) or diff > self.autosave_interval * 5
        return self._pending >= self._MAX_PENDING

    def __popup(self, widget, library):
        songs = self.get_selected_songs()
        if not songs:
            return None

        menu = SongsMenu(
            library, songs, queue=False, remove=False, delete=False, ratings=False
        )
        menu.preseparate()
        remove = MenuItem(_("_Remove"), Icons.LIST_REMOVE)
        qltk.add_fake_accel(remove, "Delete")
        remove.connect("activate", self.__remove)
        menu.prepend(remove)
        menu.show_all()
        return self.popup_menu(menu, 0, Gtk.get_current_event_time())

    def __remove(self, *args):
        self.remove_selection()



# Copyright 2006 Lukas Lalinsky
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import struct

import mutagen.asf

from quodlibet.util.path import get_temp_cover_file

from ._audio import AudioFile
from ._image import EmbeddedImage, APICType
from ._misc import AudioFileError, translate_errors


class WMAFile(AudioFile):
    mimes = [
        "audio/x-ms-wma",
        "audio/x-ms-wmv",
        "video/x-ms-asf",
        "audio/x-wma",
        "video/x-wmv",
    ]
    format = "ASF"

    # http://msdn.microsoft.com/en-us/library/dd743066%28VS.85%29.aspx
    # http://msdn.microsoft.com/en-us/library/dd743063%28VS.85%29.aspx
    # http://msdn.microsoft.com/en-us/library/dd743220%28VS.85%29.aspx
    __translate = {
        "WM/AlbumTitle": "album",
        "Title": "title",
        "Author": "artist",
        "WM/AlbumArtist": "albumartist",
        "WM/Composer": "composer",
        "WM/Writer": "lyricist",
        "WM/Conductor": "conductor",
        "WM/ModifiedBy": "remixer",
        "WM/Producer": "producer",
        "WM/ContentGroupDescription": "grouping",
        "WM/SubTitle": "discsubtitle",
        "WM/TrackNumber": "tracknumber",
        "WM/PartOfSet": "discnumber",
        "WM/BeatsPerMinute": "bpm",
        "Copyright": "copyright",
        "WM/ISRC": "isrc",
        "WM/Mood": "mood",
        "WM/EncodedBy": "encodedby",
        "MusicBrainz/Track Id": "musicbrainz_trackid",
        "MusicBrainz/Release Track Id": "musicbrainz_releasetrackid",
        "MusicBrainz/Album Id": "musicbrainz_albumid",
        "MusicBrainz/Artist Id": "musicbrainz_artistid",
        "MusicBrainz/Album Artist Id": "musicbrainz_albumartistid",
        "MusicBrainz/TRM Id": "musicbrainz_trmid",
        "MusicIP/PUID": "musicip_puid",
        "MusicBrainz/Release Group Id": "musicbrainz_releasegroupid",
        "WM/Year": "date",
        "WM/OriginalArtist": "originalartist",
        "WM/OriginalAlbumTitle": "originalalbum",
        "WM/AlbumSortOrder": "albumsort",
        "WM/ArtistSortOrder": "artistsort",
        "WM/AlbumArtistSortOrder": "albumartistsort",
        "WM/Genre": "genre",
        "WM/Publisher": "publisher",
        "WM/AuthorURL": "website",
        "Description": "comment",
    }
    __rtranslate = {v: k for k, v in __translate.items()}

    # http://msdn.microsoft.com/en-us/library/dd743065.aspx
    # note: not all names here are used by QL
    __multi_value_attr = {
        "Author",
        "WM/AlbumArtist",
        "WM/AlbumCoverURL",
        "WM/Category",
        "WM/Composer",
        "WM/Conductor",
        "WM/Director",
        "WM/Genre",
        "WM/GenreID",
        "WM/Language",
        "WM/Lyrics_Synchronised",
        "WM/Mood",
        "WM/Picture",
        "WM/Producer",
        "WM/PromotionURL",
        "WM/UserWebURL",
        "WM/Writer",
    }

    __multi_value_keys = set()
    for k, v in __translate.items():
        if k in __multi_value_attr:
            __multi_value_keys.add(v)

    def __init__(self, filename, audio=None):
        if audio is None:
            with translate_errors():
                audio = mutagen.asf.ASF(filename)
        info = audio.info

        self["~#length"] = info.length
        self["~#bitrate"] = int(info.bitrate / 1000)
        if info.channels:
            self["~#channels"] = info.channels
        self["~#samplerate"] = info.sample_rate

        type_, name, desc = info.codec_type, info.codec_name, info.codec_description

        if type_:
            self["~codec"] = type_
        encoding = "\n".join(filter(None, [name, desc]))
        if encoding:
            self["~encoding"] = encoding

        for name, values in audio.tags.items():
            if name == "WM/Picture":
                self.has_images = True
            try:
                name = self.__translate[name]
            except KeyError:
                continue
            self[name] = "\n".join(map(str, values))
        self.sanitize(filename)

    def write(self):
        with translate_errors():
            audio = mutagen.asf.ASF(self["~filename"])
        for key in self.__translate.keys():
            try:
                del audio[key]
            except KeyError:
                pass

        for key in self.realkeys():
            try:
                name = self.__rtranslate[key]
            except KeyError:
                continue
            audio.tags[name] = self.list(key)
        with translate_errors():
            audio.save()
        self.sanitize()

    def can_multiple_values(self, key=None):
        if key is None:
            return self.__multi_value_keys
        return key in self.__multi_value_keys

    def can_change(self, key=None):
        ok = self.__rtranslate.keys()
        if key is None:
            return ok
        return super().can_change(key) and (key in ok)

    def get_images(self):
        images = []

        try:
            tag = mutagen.asf.ASF(self["~filename"])
        except Exception:
            return images

        for image in tag.get("WM/Picture", []):
            try:
                (mime, desc, data, type_) = unpack_image(image.value)
            except ValueError:
                continue
            f = get_temp_cover_file(data, mime)
            images.append(EmbeddedImage(f, mime, type_=type_))

        images.sort(key=lambda c: c.sort_key)
        return images

    def get_primary_image(self):
        """Returns the primary embedded image or None"""

        try:
            tag = mutagen.asf.ASF(self["~filename"])
        except Exception:
            return None

        for image in tag.get("WM/Picture", []):
            try:
                (mime, desc, data, type_) = unpack_image(image.value)
            except ValueError:
                continue
            if type_ == APICType.COVER_FRONT:  # Only cover images
                f = get_temp_cover_file(data, mime)
                return EmbeddedImage(f, mime, type_=type_)
        return None

    can_change_images = True

    def clear_images(self):
        """Delete all embedded images"""

        with translate_errors():
            tag = mutagen.asf.ASF(self["~filename"])
            tag.pop("WM/Picture", None)
            tag.save()

        self.has_images = False

    def set_image(self, image):
        """Replaces all embedded images by the passed image"""

        with translate_errors():
            tag = mutagen.asf.ASF(self["~filename"])

        try:
            imagedata = image.read()
        except OSError as e:
            raise AudioFileError(e) from e

        # thumbnail gets used in WMP..
        data = pack_image(image.mime_type, "thumbnail", imagedata, APICType.COVER_FRONT)

        value = mutagen.asf.ASFValue(data, mutagen.asf.BYTEARRAY)
        tag["WM/Picture"] = [value]

        with translate_errors():
            tag.save()

        self.has_images = True


def unpack_image(data):
    """
    Helper function to unpack image data from a WM/Picture tag.

    The data has the following format:
    1 byte: Picture type (0-20), see ID3 APIC frame specification at
    http://www.id3.org/id3v2.4.0-frames
    4 bytes: Picture data length in LE format
    MIME type, null terminated UTF-16-LE string
    Description, null terminated UTF-16-LE string
    The image data in the given length
    """

    try:
        (type_, size) = struct.unpack_from("<bi", data)
    except struct.error as e:
        raise ValueError(e) from e
    data = data[5:]

    mime = b""
    while data:
        char, data = data[:2], data[2:]
        if char == b"\x00\x00":
            break
        mime += char
    else:
        raise ValueError("mime: missing data")

    mime = mime.decode("utf-16-le")

    description = b""
    while data:
        char, data = data[:2], data[2:]
        if char == b"\x00\x00":
            break
        description += char
    else:
        raise ValueError("desc: missing data")

    description = description.decode("utf-16-le")

    if size != len(data):
        raise ValueError("image data size mismatch")

    return (mime, description, data, type_)


def pack_image(mime, description, imagedata, type_):
    assert APICType.is_valid(type_)

    size = len(imagedata)
    data = struct.pack("<bi", type_, size)
    data += mime.encode("utf-16-le") + b"\x00\x00"
    data += description.encode("utf-16-le") + b"\x00\x00"
    data += imagedata

    return data


loader = WMAFile
types = [WMAFile]
extensions = [".wma", ".asf", ".wmv"]



# Copyright 2004-2005 Joe Wreschnig
#                2016 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os

from quodlibet import _
from quodlibet import util, qltk
from quodlibet.plugins.songshelpers import each_song, is_a_file
from quodlibet.plugins.songsmenu import SongsMenuPlugin
from quodlibet.qltk import Icons
from quodlibet.qltk.wlw import WaitLoadWindow


class IFPUpload(SongsMenuPlugin):
    PLUGIN_ID = "Send to iFP"
    PLUGIN_NAME = _("Send to iFP")
    PLUGIN_DESC = _("Uploads songs to an iRiver iFP device.")
    PLUGIN_ICON = Icons.MULTIMEDIA_PLAYER

    plugin_handles = each_song(is_a_file)

    def plugin_songs(self, songs):
        if os.system("ifp typestring"):
            qltk.ErrorMessage(
                None,
                _("No iFP device found"),
                _(
                    "Unable to contact your iFP device. Check "
                    "that the device is powered on and plugged "
                    "in, and that you have ifp-line "
                    "(http://ifp-driver.sf.net) installed."
                ),
            ).run()
            return True
        self.__madedir = []

        w = WaitLoadWindow(None, len(songs), _("Uploading %(current)d/%(total)d"))
        w.show()

        for _i, song in enumerate(songs):
            if self.__upload(song) or w.step():
                w.destroy()
                return True
        else:
            w.destroy()
        return None

    def __upload(self, song):
        filename = song["~filename"]
        basename = song("~basename")
        dirname = os.path.basename(os.path.dirname(filename))
        target = os.path.join(dirname, basename)

        # Avoid spurious calls to ifp mkdir; this can take a long time
        # on a noisy USB line.
        if dirname not in self.__madedir:
            os.system(f"ifp mkdir {dirname!r}> /dev/null 2>/dev/null")
            self.__madedir.append(dirname)
        if os.system(f"ifp upload {filename!r} {target!r} > /dev/null"):
            tmpl = _(
                "Unable to upload %s.The device may be out of space, or turned off."
            )
            qltk.ErrorMessage(
                None,
                _("Error uploading"),
                tmpl % util.bold(filename),
                escape_desc=False,
            ).run()
            return True
        return None



# Copyright 2006 Federico Pelloni <federico.pelloni@gmail.com>
#           2013 Christoph Reiter
#           2017 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import GLib
from gi.repository import Gio

from quodlibet import commands, app
from quodlibet.util import dbusutils
from quodlibet.query import Query
from quodlibet.qltk.songlist import SongList
from quodlibet.formats import decode_value


class DBusHandler:
    """
    <!DOCTYPE node PUBLIC
      "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
      "http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
    <node name="/net/sacredchao/QuodLibet">
      <interface name="org.freedesktop.DBus.Introspectable">
        <method name="Introspect">
          <arg direction="out" type="s" />
        </method>
      </interface>
      <interface name="net.sacredchao.QuodLibet">
        <signal name="SongStarted">
          <arg type="v" name="song" />
        </signal>
        <signal name="SongEnded">
          <arg type="v" name="song" />
          <arg type="v" name="skipped" />
        </signal>
        <signal name="Paused"></signal>
        <signal name="Unpaused"></signal>
        <method name="GetPosition">
          <arg direction="out" type="u" />
        </method>
        <method name="IsPlaying">
          <arg direction="out" type="b" />
        </method>
        <method name="CurrentSong">
          <arg direction="out" type="a{ss}" />
        </method>
        <method name="Rate">
          <arg direction="in"  type="s" name="rating" />
        </method>
        <method name="Next"></method>
        <method name="Previous"></method>
        <method name="Pause"></method>
        <method name="Play"></method>
        <method name="PlayPause">
          <arg direction="out" type="b" />
        </method>
        <method name="Query">
          <arg direction="in"  type="s" name="text" />
          <arg direction="out" type="aa{ss}" />
        </method>
      </interface>
    </node>
    """

    BUS_NAME = "net.sacredchao.QuodLibet"
    PATH = "/net/sacredchao/QuodLibet"

    def __init__(self, player, library):
        try:
            self._registered_ids = []
            self._method_outargs = {}
            self.library = library
            self.conn = Gio.bus_get_sync(Gio.BusType.SESSION, None)
            Gio.bus_own_name_on_connection(
                self.conn,
                self.BUS_NAME,
                Gio.BusNameOwnerFlags.NONE,
                None,
                self.__on_name_lost,
            )
            self.__register_ifaces()
        except GLib.Error:
            pass
        else:
            player.connect("song-started", self.__song_started)
            player.connect("song-ended", self.__song_ended)
            player.connect("paused", lambda player: self.Paused())
            player.connect("unpaused", lambda player: self.Unpaused())
            self._player = player

    def __on_name_lost(self, connection, name):
        for _id in self._registered_ids:
            connection.unregister_object(_id)

    def __register_ifaces(self):
        info = Gio.DBusNodeInfo.new_for_xml(self.__doc__)
        for interface in info.interfaces:
            for method in interface.methods:
                self._method_outargs[method.name] = "({})".format(
                    "".join([arg.signature for arg in method.out_args])
                )

            _id = self.conn.register_object(
                object_path=self.PATH,
                interface_info=interface,
                method_call_closure=self.__on_method_call,
            )
            self._registered_ids.append(_id)

    def __on_method_call(
        self,
        connection,
        sender,
        object_path,
        interface_name,
        method_name,
        parameters,
        invocation,
    ):
        args = list(parameters.unpack())
        result = getattr(self, method_name)(*args)
        if not isinstance(result, tuple):
            result = (result,)

        out_args = self._method_outargs[method_name]
        if out_args != "()":
            variant = GLib.Variant(out_args, result)
            invocation.return_value(variant)
        else:
            invocation.return_value(None)

    @staticmethod
    def __dict(song):
        dict = {}
        for key, value in (song or {}).items():
            value = decode_value(key, value)
            dict[key] = dbusutils.dbus_unicode_validate(value)
        if song:
            dict["~uri"] = song("~uri")
        return dict

    def __song_started(self, player, song):
        if song is not None:
            song = self.__dict(song)
            self.SongStarted(song)

    def __song_ended(self, player, song, skipped):
        if song is not None:
            song = self.__dict(song)
            self.SongEnded(song, skipped)

    def Introspect(self):
        return self.__doc__

    def SongStarted(self, song):
        self.conn.emit_signal(
            None,
            self.PATH,
            "net.sacredchao.QuodLibet",
            "SongStarted",
            GLib.Variant("(a{ss})", (song,)),
        )

    def SongEnded(self, song, skipped):
        self.conn.emit_signal(
            None,
            self.PATH,
            "net.sacredchao.QuodLibet",
            "SongEnded",
            GLib.Variant("(a{ss}b)", (song, skipped)),
        )

    def Paused(self):
        self.conn.emit_signal(
            None, self.PATH, "net.sacredchao.QuodLibet", "Paused", None
        )

    def Unpaused(self):
        self.conn.emit_signal(
            None, self.PATH, "net.sacredchao.QuodLibet", "Unpaused", None
        )

    def GetPosition(self):
        return self._player.get_position()

    def IsPlaying(self):
        return not self._player.paused

    def CurrentSong(self):
        return self.__dict(self._player.song)

    def Rate(self, rating: str):
        """
        Rates the currently playing song

        :param rating: A value between 0 and 1.
                       Prefix with +/- to indicate a relative change
        """
        commands._rating(app, rating)

    def Next(self):
        self._player.next()

    def Previous(self):
        self._player.previous()

    def Pause(self):
        self._player.paused = True

    def Play(self):
        self._player.play()

    def PlayPause(self):
        self._player.playpause()
        return self._player.paused

    def Query(self, text):
        if text is not None:
            query = Query(text, star=SongList.star)
            if query.is_parsable:
                return [
                    self.__dict(s) for s in self.library.values() if query.search(s)
                ]
        return None



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys
import subprocess

from quodlibet.qltk import Icons

if os.name == "nt" or sys.platform == "darwin":
    from quodlibet.plugins import PluginNotSupportedError

    raise PluginNotSupportedError

from gi.repository import Gtk, GObject

from quodlibet import _
from quodlibet.plugins.editing import RenameFilesPlugin
from quodlibet.util.path import iscommand
from quodlibet.util import connect_obj


class Kakasi(RenameFilesPlugin, Gtk.CheckButton):
    PLUGIN_ID = "Kana/Kanji Simple Inverter"
    PLUGIN_NAME = _("Kana/Kanji Simple Inverter")
    PLUGIN_DESC = _("Converts kana/kanji to romaji before renaming.")
    PLUGIN_ICON = Icons.EDIT_FIND_REPLACE

    __gsignals__ = {"preview": (GObject.SignalFlags.RUN_LAST, None, ())}

    def __init__(self):
        super().__init__(_("Romanize _Japanese text"), use_underline=True)
        connect_obj(self, "toggled", self.emit, "preview")

    @property
    def active(self):
        return self.get_active()

    # Use filter list rather than filter to avoid starting a new process
    # for each filename.
    def filter_list(self, originals, values):
        value = "\n".join(values)
        try:
            data = value.encode("shift-jis", "replace")
        except UnicodeEncodeError:
            return values

        proc = subprocess.Popen(
            ["kakasi", "-isjis", "-osjis", "-Ha", "-Ka", "-Ja", "-Ea", "-ka", "-s"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        result = proc.communicate(data)[0]

        try:
            return result.decode("shift-jis").strip().split("\n")
        except Exception:
            return values


if not iscommand("kakasi"):
    from quodlibet import plugins

    raise plugins.PluginImportError(
        _("Couldn't find the 'Kanji Kana Simple Inverter' (kakasi).")
    )



# Copyright 2013 Simonas Kazlauskas
#      2015-2020 Nick Boultbee
#           2019 Joschua Gandert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import glob
import os.path
import re
import sre_constants

from senf import fsn2text

from quodlibet import _
from quodlibet.plugins.cover import CoverSourcePlugin
from quodlibet.util.dprint import print_w, print_d
from quodlibet import config


def get_ext(s):
    return os.path.splitext(s)[1].lstrip(".")


def prefer_embedded():
    return config.getboolean("albumart", "prefer_embedded", False)


def word_regex(s: str) -> re.Pattern:
    return re.compile(r"(\b|_)" + s + r"(\b|_)")


class EmbeddedCover(CoverSourcePlugin):
    PLUGIN_ID = "embed-cover"
    PLUGIN_NAME = _("Embedded album covers")
    PLUGIN_DESC = _("Uses covers embedded into audio files.")

    embedded = True

    @classmethod
    def group_by(cls, song):
        # one group per song
        return song.key

    @staticmethod
    def priority():
        return 0.85 if prefer_embedded() else 0.7

    @property
    def cover(self):
        if self.song.has_images:
            image = self.song.get_primary_image()
            return image.file if image else None
        return None


class FilesystemCover(CoverSourcePlugin):
    PLUGIN_ID = "filesystem-cover"
    PLUGIN_NAME = _("Filesystem cover")
    PLUGIN_DESC = _(
        "Uses commonly named images found in common directories alongside the song."
    )
    DEBUG = False

    cover_subdirs = {"scan", "scans", "images", "covers", "artwork"}
    cover_exts = {"jpg", "jpeg", "png", "gif"}

    cover_name_regexes = {word_regex(s) for s in ("^folder$", "^cover$", "^front$")}
    cover_positive_regexes = {
        word_regex(s)
        for s in [".+front", "frontcover", "jacket", "albumart", "edited", ".+cover"]
    }
    cover_negative_regexes = {
        word_regex(s) for s in ["back", "inlay", "inset", "inside"]
    }

    @classmethod
    def group_by(cls, song):
        # in the common case this means we only search once per album
        return song("~dirname"), song.album_key

    @property
    def name(self):
        return "Filesystem"

    def __str__(self):
        return f"Filesystem in {self.group_by(self.song)[0]}"

    @staticmethod
    def priority():
        return 0.80

    @property
    def cover(self):
        # TODO: Deserves some refactoring
        if not self.song.is_file:
            return None
        print_d(f"Searching for local cover for {self.song('~filename')}")
        base = self.song("~dirname")
        images = []

        def safe_glob(*args, **kwargs):
            try:
                return glob.glob(*args, **kwargs)
            except sre_constants.error:
                # https://github.com/python/cpython/issues/89973
                # old glob would fail with invalid ranges, newer one
                # handles it correctly.
                return []

        if config.getboolean("albumart", "force_filename"):
            score = 100
            for filename in config.get("albumart", "filename").split(","):
                # Remove white space to avoid confusion (e.g. "name, name2")
                filename = filename.strip()

                escaped_path = os.path.join(glob.escape(base), filename)
                for path in safe_glob(escaped_path):
                    images.append((score, path))

                # So names and patterns at the start are preferred
                score -= 1

        if not images:
            entries = []
            try:
                entries = os.listdir(base)
            except OSError:
                print_w(f"Can't list album art directory {base}")

            fns = []
            for entry in entries:
                lentry = entry.lower()
                if get_ext(lentry) in self.cover_exts:
                    fns.append((None, entry))
                if lentry in self.cover_subdirs:
                    subdir = os.path.join(base, entry)
                    sub_entries = []
                    try:
                        sub_entries = os.listdir(subdir)
                    except OSError:
                        pass
                    for sub_entry in sub_entries:
                        lsub_entry = sub_entry.lower()
                        if get_ext(lsub_entry) in self.cover_exts:
                            fns.append((entry, sub_entry))

            for sub, fn in fns:
                dec_lfn = os.path.splitext(fsn2text(fn))[0].lower()

                score = 0
                # check for the album label number
                labelid = self.song.get("labelid", "").lower()
                if labelid and labelid in dec_lfn:
                    score += 20

                # Track-related keywords
                values = set(self.song.list("~people")) | {self.song("album")}
                lowers = [value.lower().strip() for value in values if len(value) > 1]
                total_terms = sum(len(s.split()) for s in lowers)
                total_words = len([word for word in dec_lfn.split() if len(word) > 1])
                # Penalise for many extra words in filename (wrong file?)
                length_penalty = (
                    -int((total_words - 1) / total_terms) if total_terms else 0
                )

                # Matching tag values are very good
                score += 3 * sum([value in dec_lfn for value in lowers])

                # Well known names matching exactly (folder.jpg)
                score += 4 * sum(
                    r.search(dec_lfn) is not None for r in self.cover_name_regexes
                )

                # Generic keywords
                score += 2 * sum(
                    r.search(dec_lfn) is not None for r in self.cover_positive_regexes
                )

                score -= 3 * sum(
                    r.search(dec_lfn) is not None for r in self.cover_negative_regexes
                )

                sub_text = f" (in {sub!r})" if sub else ""
                if self.DEBUG:
                    print(
                        f"[{self.song('~~people~title')}]: "
                        f"Album art {fn!r}{sub_text} "
                        f"scores {score} ({length_penalty})"
                    )
                score += length_penalty

                # Let's only match if we're quite sure.
                # This allows other sources to kick in
                if score > 2:
                    if sub is not None:
                        fn = os.path.join(sub, fn)
                    images.append((score, os.path.join(base, fn)))

        images.sort(reverse=True)
        for _score, path in images:
            # could be a directory
            if not os.path.isfile(path):
                continue
            try:
                return open(path, "rb")
            except OSError:
                print_w(f'Failed reading album art "{path}"')

        return None



# Copyright 2014,2016 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import threading
import ctypes

if os.name == "nt":
    from . import winapi

from gi.repository import GLib


def write_pipe(pipe_name, data):
    """Writes the data to the pipe or raises EnvironmentError"""

    assert isinstance(data, bytes)

    # XXX: otherwise many consecutive open fail, no idea..
    pipe_exists(pipe_name)

    filename = NamedPipeServer._get_filename(pipe_name)
    with open(filename, "wb") as h:
        h.write(data)


def pipe_exists(pipe_name):
    """Returns True if the named pipe named 'pipe_name' currently exists"""

    timeout_ms = 1
    filename = NamedPipeServer._get_filename(pipe_name)

    try:
        if winapi.WaitNamedPipeW(filename, timeout_ms) == 0:
            raise ctypes.WinError()
    except OSError:
        return False
    return True


class NamedPipeServerError(Exception):
    pass


class NamedPipeServer(threading.Thread):
    """A named pipe for Windows.

    * server:
        server = NamedPipeServer("foo", lambda data: ...)
        server.start()
        glib_loop()
        server.stop()

    * client:
        with open(NamedPipeServer.get_filename("foo"), "wb") as h:
            h.write("Hello World")

    """

    def __init__(self, name, callback):
        """name is the name of the pipe file (should be unique I guess)
        callback will be called with new data until close() is called.
        """

        super().__init__()
        self._event = threading.Event()
        self._filename = self._get_filename(name)
        self._callback = callback
        self._stopped = False

    @classmethod
    def _get_filename(cls, name):
        return f"\\\\.\\pipe\\{name}"

    def _process(self, data):
        def idle_process(data):
            if not self._stopped:
                self._callback(data)
            return False

        GLib.idle_add(idle_process, data)

    def start(self):
        super().start()
        # make sure we can use write_pipe() immediately after this returns
        self._event.wait()
        if self._stopped:
            # something went wrong (maybe another instance is running)
            raise NamedPipeServerError("Setting up named pipe failed")

    def run(self):
        buffer_size = 4096

        try:
            handle = winapi.CreateNamedPipeW(
                self._filename,
                (winapi.PIPE_ACCESS_INBOUND | winapi.FILE_FLAG_FIRST_PIPE_INSTANCE),
                (
                    winapi.PIPE_TYPE_BYTE
                    | winapi.PIPE_READMODE_BYTE
                    | winapi.PIPE_WAIT
                    | winapi.PIPE_REJECT_REMOTE_CLIENTS
                ),
                winapi.PIPE_UNLIMITED_INSTANCES,
                buffer_size,
                buffer_size,
                winapi.NMPWAIT_USE_DEFAULT_WAIT,
                None,
            )

            if handle == winapi.INVALID_HANDLE_VALUE:
                raise ctypes.WinError()

        except OSError:
            # due to FILE_FLAG_FIRST_PIPE_INSTANCE and not the first instance
            self._stopped = True
            self._event.set()
            return

        self._event.set()

        while 1:
            data = bytearray()
            try:
                if winapi.ConnectNamedPipe(handle, None) == 0:
                    raise ctypes.WinError()

                while 1:
                    readbuf = ctypes.create_string_buffer(buffer_size)
                    bytesread = winapi.DWORD()
                    try:
                        if (
                            winapi.ReadFile(
                                handle,
                                readbuf,
                                buffer_size,
                                ctypes.byref(bytesread),
                                None,
                            )
                            == 0
                        ):
                            raise ctypes.WinError()
                    except OSError:
                        break
                    else:
                        message = readbuf[: bytesread.value]

                    data += message

                if winapi.DisconnectNamedPipe(handle) == 0:
                    raise ctypes.WinError()
            except OSError:
                # better not loop forever..
                break
            finally:
                if self._stopped:
                    break  # noqa
                if data:
                    self._process(bytes(data))

        # ignore errors here..
        winapi.CloseHandle(handle)

    def stop(self):
        """After this returns the callback will no longer be called.
        Can be called multiple times.
        """

        self._event.wait()
        if self._stopped:
            return

        self._stopped = True
        try:
            with open(self._filename, "wb") as h:
                h.write(b"stop!")
        except OSError:
            pass

        self._callback = None

        self.join()



# Copyright 2006 Joe Wreschnig, Alexandre Passos
#           2014 Christoph Reiter
#           2016 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Manage a pool of routines using Python iterators."""

from gi.repository import GLib


class _Routine:
    def __init__(self, pool, func, funcid, priority, timeout, args, kwargs):
        self.priority = priority
        self.timeout = timeout
        self._source_id = None

        def wrap(func, funcid, args, kwargs):
            for _value in func(*args, **kwargs):
                yield True
            pool.remove(funcid)
            yield False

        f = wrap(func, funcid, args, kwargs)
        self.source_func = f.__next__

    @property
    def paused(self):
        """If the routine is currently running"""

        return self._source_id is None

    def step(self):
        """Raises StopIteration if the routine has nothing more to do"""

        return self.source_func()

    def resume(self):
        """Resume, if already running do nothing"""

        if not self.paused:
            return

        if self.timeout:
            self._source_id = GLib.timeout_add(
                self.timeout, self.source_func, priority=self.priority
            )
        else:
            self._source_id = GLib.idle_add(self.source_func, priority=self.priority)

    def pause(self):
        """Pause, if already paused, do nothing"""

        if self.paused:
            return

        GLib.source_remove(self._source_id)
        self._source_id = None


class CoPool:
    def __init__(self):
        self.__routines = {}

    def add(self, func, *args, **kwargs):
        """Register a routine to run in GLib main loop.

        func should be a function that returns a Python iterator (e.g.
        generator) that provides values until it should stop being called.

        Optional Keyword Arguments:
        priority -- priority to run at (default GLib.PRIORITY_LOW)
        funcid -- mutex/removal identifier for this function
        timeout -- use timeout_add (with given timeout) instead of idle_add
                   (in milliseconds)

        Only one function with the same funcid can be running at once.
        Starting a new function with the same ID will stop the old one. If
        no funcid is given, the function itself is used. The funcid must
        be usable as a hash key.
        """

        funcid = kwargs.pop("funcid", func)
        if funcid in self.__routines:
            remove(funcid)

        priority = kwargs.pop("priority", GLib.PRIORITY_LOW)
        timeout = kwargs.pop("timeout", None)

        routine = _Routine(self, func, funcid, priority, timeout, args, kwargs)
        self.__routines[funcid] = routine
        routine.resume()

    def _get(self, funcid):
        if funcid in self.__routines:
            return self.__routines[funcid]
        raise ValueError(f"no pooled routine {funcid!r}")

    def remove(self, funcid):
        """Stop a registered routine."""

        routine = self._get(funcid)
        routine.pause()
        del self.__routines[funcid]

    def remove_all(self):
        """Stop all running routines."""

        for funcid in list(self.__routines.keys()):
            self.remove(funcid)

    def pause(self, funcid):
        """Temporarily pause a registered routine."""

        routine = self._get(funcid)
        routine.pause()

    def pause_all(self):
        """Temporarily pause all registered routines."""

        for funcid in self.__routines.keys():
            self.pause(funcid)

    def resume(self, funcid):
        """Resume a paused routine."""

        routine = self._get(funcid)
        routine.resume()

    def step(self, funcid):
        """Force this function to iterate once."""

        routine = self._get(funcid)
        return routine.step()


# global instance

_copool = CoPool()

add = _copool.add
pause = _copool.pause
pause_all = _copool.pause_all
remove = _copool.remove
remove_all = _copool.remove_all
resume = _copool.resume
step = _copool.step



# Copyright 2010 Steven Robertson
#           2012 Christoph Reiter
#           2017 Nick Boultbee
#           2018 Olli Helin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk, Gdk

from quodlibet import _
from quodlibet import app
from quodlibet import config
from quodlibet.qltk import Button, Icons
from quodlibet.plugins.events import EventPlugin
from quodlibet.util import print_e, print_w

import ast

# Presets (roughly) taken from Pulseaudio equalizer
PRESET_BANDS = [
    50,
    100,
    156,
    220,
    311,
    440,
    622,
    880,
    1250,
    1750,
    2500,
    3500,
    5000,
    10000,
    20000,
]
PRESETS = {
    "flat": (_("Flat"), [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
    "live": (
        _("Live"),
        [-9.0, -5.5, 0.0, 1.5, 2.0, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.0, 1.5, 2.0],
    ),
    "full_bass_treble": (
        _("Full Bass & Treble"),
        [
            5.0,
            5.0,
            3.5,
            2.5,
            0.0,
            -7.0,
            -14.0,
            -10.0,
            -10.0,
            -8.0,
            1.0,
            1.0,
            5.0,
            7.5,
            9.5,
        ],
    ),
    "club": (
        _("Club"),
        [0.0, 0.0, 0.0, 0.0, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 2.5, 2.5, 0.0, 0.0],
    ),
    "large_hall": (
        _("Large Hall"),
        [
            7.0,
            7.0,
            7.0,
            3.5,
            3.0,
            3.0,
            3.0,
            1.5,
            0.0,
            -2.0,
            -3.5,
            -6.0,
            -9.0,
            -1.0,
            0.0,
        ],
    ),
    "party": (
        _("Party"),
        [5.0, 5.0, 5.0, 3.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.5, 5.0],
    ),
    "rock": (
        _("Rock"),
        [
            5.5,
            2.5,
            2.5,
            -8.5,
            -10.5,
            -11.0,
            -16.0,
            -14.5,
            -6.5,
            -5.5,
            -3.0,
            3.0,
            6.5,
            7.0,
            7.0,
        ],
    ),
    "soft": (
        _("Soft"),
        [3.0, 3.0, 1.0, 1.0, 0.0, -2.5, -5.0, 1.5, 0.0, 1.0, 3.0, 3.0, 6.0, 8.0, 8.0],
    ),
    "full_bass": (
        _("Full Bass"),
        [
            -16.0,
            -16.0,
            6.5,
            6.5,
            6.0,
            5.5,
            4.5,
            1.0,
            1.0,
            1.0,
            -8.0,
            -10.0,
            -16.0,
            -16.0,
            -20.5,
        ],
    ),
    "classical": (
        _("Classical"),
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -21.0, -21.0, -27.0],
    ),
    "reggae": (
        _("Reggae"),
        [0.0, 0.0, 0.0, 0.0, 0.0, -4.5, -10.0, -6.0, 0.5, 1.0, 2.0, 4.0, 4.0, 0.0, 0.0],
    ),
    "headphones": (
        _("Headphones"),
        [
            3.0,
            3.0,
            7.0,
            7.0,
            3.0,
            -1.0,
            -6.5,
            -6.0,
            -4.5,
            -4.0,
            1.0,
            1.0,
            6.0,
            8.0,
            9.0,
        ],
    ),
    "soft_rock": (
        _("Soft Rock"),
        [
            3.0,
            3.0,
            3.0,
            1.5,
            1.5,
            1.5,
            0.0,
            -3.5,
            -8.0,
            -7.0,
            -10,
            -9.0,
            -6.5,
            1.5,
            6.0,
        ],
    ),
    "full_treble": (
        _("Full Treble"),
        [
            5.0,
            -18.5,
            -18.5,
            -18.5,
            -18.5,
            -10.0,
            -8.0,
            -6.5,
            1.5,
            1.5,
            1.5,
            8.5,
            10.5,
            10.5,
            10.5,
        ],
    ),
    "dance": (
        _("Dance"),
        [
            6.0,
            4.0,
            4.0,
            1.5,
            1.5,
            1.5,
            0.0,
            0.0,
            0.0,
            1.0,
            -10.5,
            -14.0,
            -15.0,
            -7.0,
            0.0,
        ],
    ),
    "pop": (
        _("Pop"),
        [
            -3.5,
            1.0,
            2.0,
            3.0,
            5.0,
            5.5,
            6.5,
            5.0,
            3.0,
            1.5,
            0.0,
            -2.5,
            -5.0,
            -5.0,
            -3.0,
        ],
    ),
    "techno": (
        _("Techno"),
        [
            5.0,
            4.0,
            4.0,
            3.0,
            0.0,
            -4.5,
            -10.0,
            -9.0,
            -8.0,
            -5.5,
            -1.5,
            3.0,
            6.0,
            6.0,
            6.0,
        ],
    ),
    "ska": (
        _("Ska"),
        [
            -4.5,
            -8.0,
            -9.0,
            -8.5,
            -8.0,
            -6.0,
            0.0,
            1.5,
            2.5,
            2.5,
            3.0,
            3.0,
            6.0,
            6.0,
            6.0,
        ],
    ),
    "laptop": (
        _("Laptop"),
        [-1, -1, -1, -1, -5, -10, -18, -15, -10, -5, -5, -5, -5, 0, 0],
    ),
}


def interp_bands(src_band, target_band, src_gain):
    """Linear interp from one band to another. All must be sorted."""
    gain = []
    for i, b in enumerate(target_band):
        if b in src_band:
            gain.append(src_gain[i])
            continue
        idx = sorted(src_band + [b]).index(b)
        idx = min(max(idx, 1), len(src_band) - 1)
        x1, x2 = src_band[idx - 1 : idx + 1]
        y1, y2 = src_gain[idx - 1 : idx + 1]
        g = y1 + ((y2 - y1) * (b - x1)) / float(x2 - x1)
        gain.append(min(12.0, g))
    return gain


def get_config():
    try:
        config_str = config.get("plugins", "equalizer_levels", "[]")
        config_dict = ast.literal_eval(config_str)

        if isinstance(config_dict, list):
            print_w("Converting old EQ config to new format.")
            config_dict = {"Current": config_dict}
        if not isinstance(config_dict, dict):
            raise ValueError("Saved config is of wrong type.")
        if "Current" not in config_dict.keys():
            raise ValueError("Saved config was malformed.")

        # Run through the values to check everything is of correct type.
        for key in config_dict.keys():
            [float(s) for s in config_dict[key]]

        return config_dict
    except (config.Error, ValueError) as e:
        print_e(str(e))
        return {"Current": []}


class Equalizer(EventPlugin):
    PLUGIN_ID = "Equalizer"
    PLUGIN_NAME = _("Equalizer")
    PLUGIN_DESC = _(
        "Controls the tone of your music with an equalizer.\n"
        "Click or use keys to customise levels "
        "(right-click resets the band)."
    )
    PLUGIN_ICON = Icons.AUDIO_CARD

    @property
    def player_has_eq(self):
        return hasattr(app.player, "eq_bands") and app.player.eq_bands

    def __init__(self):
        super().__init__()
        self._enabled = False
        self._config = {}

    def apply(self):
        if not self.player_has_eq:
            return
        levels = self._enabled and get_config()["Current"] or []
        lbands = len(app.player.eq_bands)
        if len(levels) != lbands:
            print_w("Number of bands didn't match current. Using flat EQ.")
            levels = [0.0] * lbands
        app.player.eq_values = levels

    def enabled(self):
        self._enabled = True
        self.apply()

    def disabled(self):
        self._enabled = False
        self.apply()

    def PluginPreferences(self, win):
        main_vbox = Gtk.VBox(spacing=12)
        if not self.player_has_eq:
            l = Gtk.Label()
            l.set_markup(_("The current backend does not support equalization."))
            main_vbox.pack_start(l, False, True, 0)
            return main_vbox

        def format_hertz(band):
            if band >= 1000:
                return _("%.1f kHz") % (band / 1000.0)
            return _("%d Hz") % band

        bands = [format_hertz(band) for band in app.player.eq_bands]
        self._config = get_config()
        levels = self._config["Current"]

        # This fixes possible old corrupt config files with extra level values.
        if len(levels) != len(bands):
            print_w("Number of bands didn't match current. Using flat EQ.")
            levels = [0.0] * len(bands)

        table = Gtk.Table(rows=len(bands), columns=3)
        table.set_col_spacings(6)

        def set_band(adj, idx):
            rounded = int(adj.get_value() * 2) / 2.0
            adj.set_value(rounded)
            levels[idx] = rounded

            self._config["Current"] = levels
            config.set("plugins", "equalizer_levels", str(self._config))
            self.apply()

        adjustments = []

        for i, band in enumerate(bands):
            # align numbers and suffixes in separate rows for great justice
            lbl = Gtk.Label(label=band.split()[0])
            lbl.set_alignment(1, 0.5)
            lbl.set_padding(0, 4)
            table.attach(lbl, 0, 1, i, i + 1, xoptions=Gtk.AttachOptions.FILL)
            lbl = Gtk.Label(label=band.split()[1])
            lbl.set_alignment(1, 0.5)
            table.attach(lbl, 1, 2, i, i + 1, xoptions=Gtk.AttachOptions.FILL)
            adj = Gtk.Adjustment.new(levels[i], -24.0, 12.0, 0.5, 3, 0)
            adj.connect("value-changed", set_band, i)
            adjustments.append(adj)
            hs = Gtk.HScale(adjustment=adj)
            hs.connect("button-press-event", self.__rightclick)
            hs.set_draw_value(True)
            hs.set_value_pos(Gtk.PositionType.RIGHT)
            hs.connect("format-value", lambda s, v: _("%.1f dB") % v)
            table.attach(hs, 2, 3, i, i + 1)
        main_vbox.pack_start(table, True, True, 0)

        # Reset EQ button
        def clicked_rb(button):
            [adj.set_value(0) for adj in adjustments]
            self._combo_default.set_active(0)
            self._combo_custom.set_active(0)

        # Delete custom preset button
        def clicked_db(button):
            selected_index = self._combo_custom.get_active()
            if selected_index < 1:
                return  # Select…
            selected = self._combo_custom.get_active_text()
            self._combo_custom.set_active(0)
            self._combo_custom.remove(selected_index)
            del self._config[selected]
            config.set("plugins", "equalizer_levels", str(self._config))

        # Save custom preset button
        def clicked_sb(button):
            name = self._preset_name_entry.get_text()
            is_new = name not in self._config.keys()

            levels = [adj.get_value() for adj in adjustments]
            self._config[name] = levels
            config.set("plugins", "equalizer_levels", str(self._config))

            self._preset_name_entry.set_text("")
            if is_new:
                self._combo_custom.append_text(name)

            def find_iter(list_store, text):
                i = list_store.get_iter_first()
                while i is not None:
                    if list_store.get_value(i, 0) == text:
                        return i
                    i = list_store.iter_next(i)
                return None

            itr = find_iter(self._combo_custom.get_model(), name)
            self._combo_custom.set_active_iter(itr)

        sorted_presets = sorted(PRESETS.items())

        def default_combo_changed(combo):
            if combo.get_active() < 1:
                return  # Select…
            self._combo_custom.set_active(0)
            gain = sorted_presets[combo.get_active() - 1][1][1]
            gain = interp_bands(PRESET_BANDS, app.player.eq_bands, gain)
            for g, a in zip(gain, adjustments, strict=False):
                a.set_value(g)

        def custom_combo_changed(combo):
            if combo.get_active() < 1:
                # Case: Select…
                self._delete_button.set_sensitive(False)
                return
            self._combo_default.set_active(0)
            self._delete_button.set_sensitive(True)
            gain = self._config[combo.get_active_text()]
            for g, a in zip(gain, adjustments, strict=False):
                a.set_value(g)

        def save_name_changed(entry):
            name = entry.get_text()
            if not name or name == "Current" or name.isspace():
                self._save_button.set_sensitive(False)
            else:
                self._save_button.set_sensitive(True)

        frame = Gtk.Frame(label=_("Default presets"), label_xalign=0.5)
        main_middle_hbox = Gtk.HBox(spacing=6)

        # Default presets
        combo = Gtk.ComboBoxText()
        self._combo_default = combo
        combo.append_text(_("Select…"))
        combo.set_active(0)
        for _key, (name, _gain) in sorted_presets:
            combo.append_text(name)
        combo.connect("changed", default_combo_changed)

        # This block is just for padding.
        padboxv = Gtk.VBox()
        padboxv.pack_start(combo, True, True, 6)
        padboxh = Gtk.HBox()
        padboxh.pack_start(padboxv, True, True, 6)
        frame.add(padboxh)

        main_middle_hbox.pack_start(frame, True, True, 0)

        reset = Button(_("_Reset EQ"), Icons.EDIT_UNDO)
        reset.connect("clicked", clicked_rb)
        main_middle_hbox.pack_start(reset, False, False, 0)

        main_vbox.pack_start(main_middle_hbox, False, False, 0)

        frame = Gtk.Frame(label=_("Custom presets"), label_xalign=0.5)
        main_bottom_vbox = Gtk.VBox()

        # Custom presets
        combo = Gtk.ComboBoxText()
        self._combo_custom = combo
        combo.append_text(_("Select…"))
        combo.set_active(0)

        custom_presets = self._config.keys() - {"Current"}
        for key in custom_presets:
            combo.append_text(key)
        combo.connect("changed", custom_combo_changed)
        hb = Gtk.HBox(spacing=6)
        hb.pack_start(combo, True, True, 0)

        delete = Button(_("_Delete selected"), Icons.EDIT_DELETE)
        delete.connect("clicked", clicked_db)
        delete.set_sensitive(False)
        self._delete_button = delete
        hb.pack_start(delete, False, False, 0)

        main_bottom_vbox.pack_start(hb, True, True, 6)
        hs = Gtk.HSeparator()
        main_bottom_vbox.pack_start(hs, True, True, 6)

        hb = Gtk.HBox()
        l = Gtk.Label(label=_("Preset name for saving:"))
        hb.pack_start(l, False, False, 0)
        main_bottom_vbox.pack_start(hb, False, False, 0)

        e = Gtk.Entry()
        e.connect("changed", save_name_changed)
        self._preset_name_entry = e
        hb = Gtk.HBox(spacing=6)
        hb.pack_start(e, True, True, 0)

        save = Button(_("_Save"), Icons.DOCUMENT_SAVE)
        save.connect("clicked", clicked_sb)
        save.set_sensitive(False)
        self._save_button = save
        hb.pack_start(save, False, False, 0)

        main_bottom_vbox.pack_start(hb, True, True, 6)

        # This block is just for padding.
        padboxh = Gtk.HBox()
        padboxh.pack_start(main_bottom_vbox, True, True, 6)
        frame.add(padboxh)

        main_vbox.pack_start(frame, True, True, 0)
        return main_vbox

    def __rightclick(self, hs, event):
        if event.button == Gdk.BUTTON_SECONDARY:
            hs.set_value(0)



# Copyright 2012-2020 Nick Boultbee
#           2012,2014,2018 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Show directories and files in the default system file browser"""

import os
import subprocess
from collections.abc import Iterable

from gi.repository import GLib
from gi.repository import Gio
from gi.repository import Gtk
from senf import fsn2uri, fsnative

from quodlibet import print_d
from quodlibet.util import is_windows, is_osx


def show_files(dirname, entries: Iterable[fsnative] | None = None):
    """Shows the directory in the default file browser and if passed
    a list of directory entries will highlight those.

    Depending on the system/platform this might highlight all files passed,
    or only one of them, or none at all.

    Args:
        dirname (fsnative): Path to the directory
        entries (List[fsnative]): List of (relative) filenames in the directory
        entries (List[fsnative]): List of (relative) filenames in the directory
    Returns:
        bool: if the action was successful or not
    """
    entries = entries or []
    assert isinstance(dirname, fsnative)
    assert all(isinstance(e, fsnative) and os.path.basename(e) == e for e in entries)

    dirname = os.path.abspath(dirname)

    if is_windows():
        implementations = [_show_files_win32]
    elif is_osx():
        implementations = [_show_files_finder]
    else:
        implementations = [
            _show_files_fdo,
            _show_files_thunar,
            _show_files_xdg_open,
            _show_files_gnome_open,
        ]

    for impl in implementations:
        try:
            impl(dirname, entries)
        except BrowseError as e:
            print_d(f"Couldn't show files with {impl} ({e}), ignoring.")
            continue
        else:
            return True
    return False


def show_songs(songs):
    """Returns False if showing any of them failed"""

    dirs = {}
    for s in songs:
        dirs.setdefault(s("~dirname"), []).append(s("~basename"))

    for dirname, entries in sorted(dirs.items()):
        status = show_files(dirname, entries)
        if not status:
            return False
    return True


class BrowseError(Exception):
    pass


def _get_startup_id():
    from quodlibet import app

    app_name = type(app.window).__name__
    return "%s_TIME%d" % (app_name, Gtk.get_current_event_time())


def _get_dbus_proxy(name, path, iface):
    bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
    return Gio.DBusProxy.new_sync(
        bus, Gio.DBusProxyFlags.NONE, None, name, path, iface, None
    )


def _show_files_fdo(dirname, entries):
    # https://www.freedesktop.org/wiki/Specifications/file-manager-interface/
    FDO_PATH = "/org/freedesktop/FileManager1"
    FDO_NAME = "org.freedesktop.FileManager1"
    FDO_IFACE = "org.freedesktop.FileManager1"

    try:
        dbus_proxy = _get_dbus_proxy(FDO_NAME, FDO_PATH, FDO_IFACE)

        if not entries:
            dbus_proxy.ShowFolders("(ass)", [fsn2uri(dirname)], _get_startup_id())
        else:
            item_uri = fsn2uri(os.path.join(dirname, entries[0]))
            dbus_proxy.ShowItems("(ass)", [item_uri], _get_startup_id())
    except GLib.Error as e:
        raise BrowseError(e) from e


def _show_files_thunar(dirname, entries):
    # https://git.xfce.org/xfce/thunar/tree/thunar/thunar-dbus-service-infos.xml
    XFCE_PATH = "/org/xfce/FileManager"
    XFCE_NAME = "org.xfce.FileManager"
    XFCE_IFACE = "org.xfce.FileManager"

    try:
        dbus_proxy = _get_dbus_proxy(XFCE_NAME, XFCE_PATH, XFCE_IFACE)

        if not entries:
            dbus_proxy.DisplayFolder("(sss)", fsn2uri(dirname), "", _get_startup_id())
        else:
            dbus_proxy.DisplayFolderAndSelect(
                "(ssss)", fsn2uri(dirname), entries[0], "", _get_startup_id()
            )
    except GLib.Error as e:
        raise BrowseError(e) from e


def _show_files_gnome_open(dirname, *args):
    try:
        if subprocess.call(["gnome-open", dirname]) != 0:
            raise OSError("gnome-open error return status")
    except OSError as e:
        raise BrowseError(e) from e


def _show_files_xdg_open(dirname, *args):
    try:
        if subprocess.call(["xdg-open", dirname]) != 0:
            raise OSError("xdg-open error return status")
    except OSError as e:
        raise BrowseError(e) from e


def _show_files_win32(dirname, entries):
    if not is_windows():
        raise BrowseError("windows only")

    if not entries:
        # open_folder_and_select_items will open the parent if no items
        # are passed, so execute explorer directly for that case
        try:
            if subprocess.call(["explorer", dirname]) != 0:
                raise OSError("explorer error return status")
        except OSError as e:
            raise BrowseError(e) from e
    else:
        from quodlibet.util.windows import open_folder_and_select_items

        try:
            open_folder_and_select_items(dirname, entries)
        except OSError as e:
            raise BrowseError(e) from e


def _show_files_finder(dirname, *args):
    if not is_osx():
        raise BrowseError("OS X only")

    try:
        if subprocess.call(["open", "-R", dirname]) != 0:
            raise OSError("open error return status")
    except OSError as e:
        raise BrowseError(e) from e



# Copyright 2006 Joe Wreschnig
#           2013 Christoph Reiter
#        2016-21 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


try:
    from collections import abc
except ImportError:
    import collections as abc  # type: ignore
from collections import defaultdict
from typing import Any

from .misc import total_ordering


@total_ordering
class DictMixin:
    """Implement the dict API using keys() and __*item__ methods.

    Similar to UserDict.DictMixin, this takes a class that defines
    __getitem__, __setitem__, __delitem__, and keys(), and turns it
    into a full dict-like object.

    UserDict.DictMixin is not suitable for this purpose because it's
    an old-style class.

    This class is not optimized for very large dictionaries; many
    functions have linear memory requirements. I recommend you
    override some of these functions if speed is required.
    """

    def __getitem__(self, key):
        raise NotImplementedError()

    def __iter__(self):
        return iter(self.keys())

    def has_key(self, key: Any) -> bool:
        try:
            self[key]
        except KeyError:
            return False
        else:
            return True

    __contains__ = has_key

    def iterkeys(self):
        return iter(self.keys())

    def values(self):
        return [self[k] for k in self.keys()]

    def itervalues(self):
        return iter(self.values())

    def items(self):
        return list(zip(self.keys(), self.values(), strict=False))

    def iteritems(self):
        return iter(self.items())

    def clear(self):
        for key in list(self.keys()):
            del self[key]

    def pop(self, key, *args):
        if len(args) > 1:
            raise TypeError("pop takes at most two arguments")
        try:
            value = self[key]
        except KeyError:
            if args:
                return args[0]
            raise
        del self[key]
        return value

    def popitem(self):
        try:
            key = list(self.keys())[0]
            return key, self.pop(key)
        except IndexError as e:
            raise KeyError("dictionary is empty") from e

    def update(self, other=None, **kwargs):
        if other is None:
            self.update(kwargs)
            other = {}

        try:
            for key, value in other.items():
                self[key] = value
        except AttributeError:
            for key, value in other:
                self[key] = value

    def setdefault(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            self[key] = default
            return default

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __repr__(self):
        return repr(dict(self.items()))

    def __eq__(self, other):
        return dict(self.items()) == other

    def __lt__(self, other):
        return dict(self.items()) < other

    __hash__ = object.__hash__

    def __len__(self):
        return len(self.keys())


class DictProxy(DictMixin):
    def __init__(self, *args, **kwargs):
        self.__dict = {}
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        return self.__dict[key]

    def __setitem__(self, key, value):
        self.__dict[key] = value

    def __delitem__(self, key):
        del self.__dict[key]

    def keys(self):
        return self.__dict.keys()


class HashedList(abc.MutableSequence):
    """A list-like collection that can only take hashable items
    and provides fast membership tests.

    Can handle duplicate entries.
    """

    def __init__(self, arg=None):
        self._map = defaultdict(int)
        if arg is None:
            self._data = []
            return

        self._data = list(arg)
        for item in arg:
            self._map[item] += 1

    def __setitem__(self, index, item):
        old_items = self._data[index]
        if not isinstance(index, slice):
            old_items = [old_items]

        for old in old_items:
            self._map[old] -= 1
            if not self._map[old]:
                del self._map[old]

        self._data[index] = item

        items = item
        if not isinstance(index, slice):
            items = [items]

        for item in items:
            self._map[item] += 1

    def __getitem__(self, index):
        return self._data[index]

    def __delitem__(self, index):
        items = self._data[index]
        if not isinstance(index, slice):
            items = [items]
        for item in items:
            self._map[item] -= 1
            if not self._map[item]:
                del self._map[item]
        del self._data[index]

    def __len__(self):
        return len(self._data)

    def insert(self, index, item):
        self._data.insert(index, item)
        self._map[item] += 1

    def __contains__(self, item):
        return item in self._map

    def __iter__(self):
        yield from self._data

    def has_duplicates(self):
        """Returns True if any item is contained more than once"""
        return len(self._map) != len(self)

    def __repr__(self):
        return repr(self._data)



# Copyright 2014 Christoph Reiter <reiter.christoph@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""A reactive tcp server"""

import socket
import errno

from gi.repository import Gio, GLib

from quodlibet.qltk import io_add_watch


class ServerError(Exception):
    pass


class BaseTCPServer:
    def __init__(self, port, connection_class, debug=False):
        """port -- IP port
        connection_class -- BaseTCPConnection subclass
        """

        self._connections = []
        self._port = port
        self._connection_class = connection_class
        self._sock_service = None
        self._debug = debug

    def log(self, msg):
        """Override for logging"""

    def start(self):
        """Start accepting connections.

        May raise ServerError.
        """

        assert not self._sock_service

        service = Gio.SocketService.new()
        try:
            service.add_inet_port(self._port, None)
        except GLib.GError as e:
            raise ServerError(e) from e
        except OverflowError as e:
            raise ServerError(f"port: {e}") from e
        self._id = service.connect("incoming", self._incoming_connection_cb)
        service.start()
        self._sock_service = service

    def stop(self):
        """Stop accepting connections and close all existing connections.

        Can be called multiple times.
        """

        if not self._sock_service:
            return

        self._sock_service.disconnect(self._id)
        if self._sock_service.is_active():
            self._sock_service.stop()
        self._sock_service = None

        for conn in list(self._connections):
            conn.close()

        assert not self._connections, self._connections

    def handle_init(self):
        """Gets called if a new connection starts and there was none before"""

    def handle_idle(self):
        """Gets called once the last connection closes"""

    def _remove_connection(self, conn):
        """Called by the connection class on close"""

        self._connections.remove(conn)
        del conn._gio_connection

        if not self._connections:
            self.handle_idle()

    def _incoming_connection_cb(self, service, connection, *args):
        try:
            addr = connection.get_remote_address()
        except GLib.GError:
            addr_string = "?.?.?.?"
        else:
            addr_string = addr.get_address().to_string()

        fd = connection.get_socket().get_fd()
        sock = socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)
        sock.setblocking(0)

        msg = "New connection from %s at socket %d" % (addr_string, sock.fileno())
        if self._debug:
            self.log(msg)

        if not self._connections:
            self.handle_init()
        tcp_conn = self._connection_class(self, sock)
        self._connections.append(tcp_conn)
        # XXX: set unneeded `connection` to keep the reference
        tcp_conn._gio_connection = connection
        tcp_conn.handle_init(self)


class BaseTCPConnection:
    """Abstract base class for TCP connections.

    Subclasses need to implement the handle_*() can_*() methods.
    """

    def __init__(self, server, sock):
        self._server = server
        self._sock = sock

        self._in_id = None
        self._out_id = None
        self._closed = False

    @property
    def name(self):
        return str(self._sock.fileno())

    def start_read(self):
        """Start to read and call handle_read() if data is available.

        Only call once.
        """

        assert self._in_id is None and not self._closed

        def can_read_cb(sock, flags, *args):
            if flags & (GLib.IOCondition.HUP | GLib.IOCondition.ERR):
                self.close()
                return False

            if flags & GLib.IOCondition.IN:
                while True:
                    try:
                        data = sock.recv(4096)
                    except OSError as e:
                        if e.errno in (errno.EWOULDBLOCK, errno.EAGAIN):
                            return True
                        if e.errno == errno.EINTR:
                            continue
                        self.close()
                        return False
                    break

                if not data:
                    self.close()
                    return False

                self.handle_read(data)
                # the implementation could close in handle_read()
                if not self._closed:
                    self.start_write()

            return True

        self._in_id = io_add_watch(
            self._sock,
            GLib.PRIORITY_DEFAULT,
            GLib.IOCondition.IN | GLib.IOCondition.ERR | GLib.IOCondition.HUP,
            can_read_cb,
        )

    def start_write(self):
        """Trigger at least one call to handle_write() if can_write is True.

        Used to start writing to a client not triggered by a client request.
        """

        assert not self._closed

        write_buffer = bytearray()

        def can_write_cb(sock, flags, *args):
            if flags & (GLib.IOCondition.HUP | GLib.IOCondition.ERR):
                self.close()
                return False

            if flags & GLib.IOCondition.OUT:
                if self.can_write():
                    write_buffer.extend(self.handle_write())
                if not write_buffer:
                    self._out_id = None
                    return False

                while True:
                    try:
                        result = sock.send(write_buffer)
                    except OSError as e:
                        if e.errno in (errno.EWOULDBLOCK, errno.EAGAIN):
                            return True
                        if e.errno == errno.EINTR:
                            continue
                        self.close()
                        return False
                    break

                del write_buffer[:result]

            return True

        if self._out_id is None:
            self._out_id = io_add_watch(
                self._sock,
                GLib.PRIORITY_DEFAULT,
                GLib.IOCondition.OUT | GLib.IOCondition.ERR | GLib.IOCondition.HUP,
                can_write_cb,
            )

    def close(self):
        """Close this connection. Can be called multiple times.

        handle_close() will be called last.
        """

        if self._closed:
            return
        self._closed = True

        if self._in_id is not None:
            GLib.source_remove(self._in_id)
            self._in_id = None
        if self._out_id is not None:
            GLib.source_remove(self._out_id)
            self._out_id = None

        self.handle_close()
        self._server._remove_connection(self)
        self._sock.close()

    def handle_init(self, server):
        """Called first, gets passed the BaseServer instance"""

        raise NotImplementedError

    def handle_read(self, data):
        """Called if new data was read"""

        raise NotImplementedError

    def handle_write(self):
        """Called if new data can be written, should return the data"""

        raise NotImplementedError

    def can_write(self):
        """Should return True if handle_write() can return data"""

        raise NotImplementedError

    def handle_close(self):
        """Called last when the connection gets closed"""

        raise NotImplementedError



# Copyright 2018 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gio, GLib

from quodlibet import print_d, print_w
from ._base import SessionClient, SessionError


class GnomeSessionClient(SessionClient):
    DBUS_NAME = "org.gnome.SessionManager"
    DBUS_OBJECT_PATH = "/org/gnome/SessionManager"
    DBUS_MAIN_INTERFACE = "org.gnome.SessionManager"
    DBUS_CLIENT_INTERFACE = "org.gnome.SessionManager.ClientPrivate"

    def __init__(self):
        super().__init__()
        self._client_priv = None
        self._client_path = None
        self._sig_id = None

    def open(self, app):
        print_d("Connecting with gnome session manager")
        try:
            bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
            session_mgr = Gio.DBusProxy.new_sync(
                bus,
                Gio.DBusProxyFlags.NONE,
                None,
                self.DBUS_NAME,
                self.DBUS_OBJECT_PATH,
                self.DBUS_MAIN_INTERFACE,
                None,
            )
            if session_mgr.get_name_owner() is None:
                raise SessionError(f"{self.DBUS_NAME} unowned")
            client_path = session_mgr.RegisterClient("(ss)", app.id, "")
            if client_path is None:
                # https://github.com/quodlibet/quodlibet/issues/2435
                raise SessionError("Broken session manager implementation, likely LXDE")

            client_priv = Gio.DBusProxy.new_sync(
                bus,
                Gio.DBusProxyFlags.NONE,
                None,
                self.DBUS_NAME,
                client_path,
                self.DBUS_CLIENT_INTERFACE,
                None,
            )

            def g_signal_cb(proxy, sender, signal, args):
                if signal == "EndSession":
                    print_d("GSM sent EndSession: going down")
                    proxy.EndSessionResponse("(bs)", True, "")
                    app.quit()
                elif signal == "Stop":
                    print_d("GSM sent Stop: going down")
                    app.quit()
                elif signal == "QueryEndSession":
                    print_d("GSM sent QueryEndSession")
                    proxy.EndSessionResponse("(bs)", True, "")

            self._sig_id = client_priv.connect("g-signal", g_signal_cb)
            self._client_priv = client_priv
            self._client_path = client_path
            print_d(f"Connected with gnome session manager: {client_path}")
        except GLib.Error as e:
            raise SessionError(e) from e

    def close(self):
        if self._client_priv is None:
            return

        self._client_priv.disconnect(self._sig_id)
        self._sig_id = None

        try:
            bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
            session_mgr = Gio.DBusProxy.new_sync(
                bus,
                Gio.DBusProxyFlags.NONE,
                None,
                self.DBUS_NAME,
                self.DBUS_OBJECT_PATH,
                self.DBUS_MAIN_INTERFACE,
                None,
            )
            session_mgr.UnregisterClient("(o)", self._client_path)
        except GLib.Error as e:
            print_w(str(e))

        print_d(f"Disconnected from gnome session manager: {self._client_path}")
        self._client_path = None



# Copyright 2005 Inigo Serna
#           2018 Phoenix Dailey, Fredrik Strupe
#           2022 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from urllib.parse import quote

from gi.repository import Gtk

from quodlibet import _
from quodlibet import config
from quodlibet import app
from quodlibet import util
from quodlibet.qltk.entry import Entry
from quodlibet.qltk.data_editors import TagListEditor
from quodlibet.qltk import Icons, get_top_parent, ErrorMessage
from quodlibet.plugins.songsmenu import SongsMenuPlugin

WIKI_URL = "https://%s.wikipedia.org/wiki/Special:Search/"


def get_lang():
    return config.get("plugins", "wiki_lang", "en")


def set_lang(value):
    config.set("plugins", "wiki_lang", value)


class WikiSearch(SongsMenuPlugin):
    PLUGIN_ID = "Search Tag in Wikipedia"
    PLUGIN_NAME = _("Search Tag in Wikipedia")
    PLUGIN_DESC = _(
        "Opens a browser window with the Wikipedia article "
        "on the selected song's corresponding tag."
    )
    PLUGIN_ICON = Icons.APPLICATION_INTERNET

    DEFAULT_TAGS = ["album", "artist", "composer"]

    @classmethod
    def changed(cls, e):
        set_lang(e.get_text())

    @classmethod
    def PluginPreferences(cls, parent):
        hb = Gtk.HBox(spacing=3)
        hb.set_border_width(6)
        e = Entry(max_length=2)
        e.set_width_chars(3)
        e.set_max_width_chars(3)
        e.set_text(get_lang())
        e.connect("changed", cls.changed)
        hb.pack_start(
            Gtk.Label(label=_("Search at %(website)s") % {"website": "https://"}),
            False,
            True,
            0,
        )
        hb.pack_start(e, False, True, 0)
        hb.pack_start(Gtk.Label(label=".wikipedia.org"), False, True, 0)
        vb = Gtk.VBox(spacing=6)
        vb.pack_start(hb, False, False, 0)

        def _open_editor(widget):
            def _editor_closed(widget):
                config.setlist("plugins", "wiki_tags", widget.tags)

            tags = config.getlist("plugins", "wiki_tags", cls.DEFAULT_TAGS)
            editor = TagListEditor(_("Edit Tags"), [] if tags == [""] else tags)
            editor.set_transient_for(get_top_parent(parent))
            editor.connect("destroy", _editor_closed)
            editor.show()

        button = Gtk.Button(_("Edit Tags"))
        button.connect("clicked", _open_editor)
        vb.pack_start(button, False, True, 0)
        vb.show_all()

        return vb

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self.selected_tag = None
        self.update_submenu()

    def update_submenu(self):
        submenu = Gtk.Menu()
        tags = config.getlist("plugins", "wiki_tags", self.DEFAULT_TAGS)
        for tag in tags:
            if tag:
                item = Gtk.MenuItem(label=util.tag(tag))
                item.connect("activate", self._set_selected_tag, tag)
                submenu.append(item)

        if submenu.get_children():
            self.set_submenu(submenu)
        else:
            self.set_sensitive(False)

    def _set_selected_tag(self, widget, tag):
        self.selected_tag = tag

    def plugin_songs(self, songs):
        if not self.selected_tag:
            return
        l = dict.fromkeys([song(self.selected_tag) for song in songs]).keys()
        # If no tags values were found, show an error dialog
        if list(l) == [""]:
            ErrorMessage(
                app.window,
                _("Search failed"),
                _('Tag "%s" not found.') % self.selected_tag,
            ).run()
            return
        for a in l:
            # Only search for non-empty tags
            if a:
                a = quote(str(a).title().replace(" ", "_"))
                util.website(WIKI_URL % get_lang() + a)



# Copyright 2019 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import sys
import unicodedata

_remove_punctuation_trans = dict.fromkeys(
    i for i in range(sys.maxunicode) if unicodedata.category(chr(i)).startswith("P")
)


def remove_punctuation(s: str) -> str:
    """Removes all known Unicode punctuation from the given string"""
    return s.translate(_remove_punctuation_trans)


def remove_diacritics(s: str) -> str:
    """Canonicalises and removes all diacritics from the given string"""
    return "".join(
        c for c in unicodedata.normalize("NFKD", str(s)) if not unicodedata.combining(c)
    )



# Copyright 2005 Alexey Bobyakov <claymore.ws@gmail.com>, Joe Wreschnig
# Copyright 2006 Lukas Lalinsky
#      2020,2025 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from mutagen.mp4 import MP4, MP4Cover

from quodlibet.util.path import get_temp_cover_file
from quodlibet.util.string import decode

from ._audio import AudioFile
from ._misc import AudioFileError, translate_errors
from ._image import EmbeddedImage

ITUNES: str = "----:com.apple.iTunes"


class MP4File(AudioFile):
    format = "MPEG-4"
    mimes = ["audio/mp4", "audio/x-m4a", "audio/mpeg4", "audio/aac"]

    __translate = {
        "\xa9nam": "title",
        "\xa9alb": "album",
        "\xa9ART": "artist",
        "aART": "albumartist",
        "\xa9wrt": "composer",
        "\xa9day": "date",
        "\xa9cmt": "comment",
        "\xa9grp": "grouping",
        "\xa9gen": "genre",
        "tmpo": "bpm",
        "\xa9too": "encodedby",  # FIXME: \xa9enc should be encodedby
        "desc": "description",  # (usually used in podcasts)
        "cprt": "copyright",
        "soal": "albumsort",
        "soaa": "albumartistsort",
        "soar": "artistsort",
        "sonm": "titlesort",
        "soco": "composersort",
        f"{ITUNES}:CONDUCTOR": "conductor",
        f"{ITUNES}:DISCSUBTITLE": "discsubtitle",
        f"{ITUNES}:LANGUAGE": "language",
        f"{ITUNES}:MOOD": "mood",
        f"{ITUNES}:MusicBrainz Artist Id": "musicbrainz_artistid",
        f"{ITUNES}:MusicBrainz Track Id": "musicbrainz_trackid",
        f"{ITUNES}:MusicBrainz Release Track Id": "musicbrainz_releasetrackid",
        f"{ITUNES}:MusicBrainz Album Id": "musicbrainz_albumid",
        f"{ITUNES}:MusicBrainz Album Artist Id": "musicbrainz_albumartistid",
        f"{ITUNES}:MusicIP PUID": "musicip_puid",
        f"{ITUNES}:MusicBrainz Album Status": "musicbrainz_albumstatus",
        f"{ITUNES}:MusicBrainz Album Type": "musicbrainz_albumtype",
        f"{ITUNES}:MusicBrainz Album Release Country": "releasecountry",
        f"{ITUNES}:MusicBrainz Release Group Id": "musicbrainz_releasegroupid",
        f"{ITUNES}:replaygain_album_gain": "replaygain_album_gain",
        f"{ITUNES}:replaygain_album_peak": "replaygain_album_peak",
        f"{ITUNES}:replaygain_track_gain": "replaygain_track_gain",
        f"{ITUNES}:replaygain_track_peak": "replaygain_track_peak",
        f"{ITUNES}:replaygain_reference_loudness": "replaygain_reference_loudness",
    }
    __rtranslate = {v: k for k, v in __translate.items()}

    __tupletranslate = {
        "disk": "discnumber",
        "trkn": "tracknumber",
    }
    __rtupletranslate = {v: k for k, v in __tupletranslate.items()}

    def __init__(self, filename):
        with translate_errors():
            audio = MP4(filename)
        self["~codec"] = audio.info.codec_description
        self["~#length"] = audio.info.length
        self["~#bitrate"] = int(audio.info.bitrate / 1000)
        if audio.info.channels:
            self["~#channels"] = audio.info.channels
        self["~#samplerate"] = audio.info.sample_rate
        self["~#bitdepth"] = audio.info.bits_per_sample

        for key, values in audio.items():
            if key in self.__tupletranslate:
                if values:
                    name = self.__tupletranslate[key]
                    cur, total = values[0]
                    if total:
                        self[name] = f"{cur:d}/{total:d}"
                    else:
                        self[name] = str(cur)
            elif key in self.__translate:
                name = self.__translate[key]
                if key == "tmpo":
                    self[name] = "\n".join(map(str, values))
                elif key.startswith("----"):
                    self[name] = "\n".join(decode(v).strip("\x00") for v in values)
                else:
                    self[name] = "\n".join(values)
            elif key == "covr":
                self.has_images = True
        self.sanitize(filename)

    def write(self):
        with translate_errors():
            audio = MP4(self["~filename"])

        for key in list(self.__translate.keys()) + list(self.__tupletranslate.keys()):
            try:
                del audio[key]
            except KeyError:
                pass

        for key in self.realkeys():
            try:
                name = self.__rtranslate[key]
            except KeyError:
                continue
            values = self.list(key)
            if name == "tmpo":
                values = [int(round(float(v))) for v in values]
            elif name.startswith("----"):
                values = [v.encode("utf-8") for v in values]
            audio[name] = values
        track, tracks = self("~#track"), self("~#tracks", 0)
        if track:
            audio["trkn"] = [(track, tracks)]
        disc, discs = self("~#disc"), self("~#discs", 0)
        if disc:
            audio["disk"] = [(disc, discs)]
        with translate_errors():
            audio.save()
        self.sanitize()

    def can_multiple_values(self, key=None):
        if key is None:
            return []
        return False

    def can_change(self, key=None):
        ok = list(self.__rtranslate.keys()) + list(self.__rtupletranslate.keys())
        if key is None:
            return ok
        return super().can_change(key) and (key in ok)

    def get_images(self):
        images = []

        try:
            tag = MP4(self["~filename"])
        except Exception:
            return []

        for cover in tag.get("covr", []):
            if cover.imageformat == MP4Cover.FORMAT_JPEG:
                mime = "image/jpeg"
            elif cover.imageformat == MP4Cover.FORMAT_PNG:
                mime = "image/png"
            else:
                mime = "image/"

            f = get_temp_cover_file(cover, mime)
            images.append(EmbeddedImage(f, mime))

        return images

    def get_primary_image(self):
        try:
            tag = MP4(self["~filename"])
        except Exception:
            return None

        for cover in tag.get("covr", []):
            if cover.imageformat == MP4Cover.FORMAT_JPEG:
                mime = "image/jpeg"
            elif cover.imageformat == MP4Cover.FORMAT_PNG:
                mime = "image/png"
            else:
                mime = "image/"

            f = get_temp_cover_file(cover, mime)
            return EmbeddedImage(f, mime)
        return None

    can_change_images = True

    def clear_images(self):
        """Delete all embedded images"""

        with translate_errors():
            tag = MP4(self["~filename"])
            tag.pop("covr", None)
            tag.save()

        self.has_images = False

    def set_image(self, image):
        """Replaces all embedded images by the passed image"""

        if image.mime_type == "image/jpeg":
            image_format = MP4Cover.FORMAT_JPEG
        elif image.mime_type == "image/png":
            image_format = MP4Cover.FORMAT_PNG
        else:
            raise AudioFileError(f"mp4: Unsupported image format {image.mime_type!r}")

        with translate_errors():
            tag = MP4(self["~filename"])

        try:
            data = image.read()
        except OSError:
            return

        cover = MP4Cover(data, image_format)
        tag["covr"] = [cover]

        with translate_errors():
            tag.save()

        self.has_images = True


loader = MP4File
types = [MP4File]
extensions = [".mp4", ".m4a", ".m4b", ".m4v", ".3gp", ".3g2", ".3gp2"]



# Copyright 2004-2005 Joe Wreschnig, Michael Urman, Iñigo Serna
#           2012,2013 Christoph Reiter
#           2010-2021 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import sys
import os

from quodlibet import _
from quodlibet.cli import process_arguments, exit_
from quodlibet.util.dprint import print_d, print_, print_exc


def main(argv=None):
    if argv is None:
        argv = sys.argv

    import quodlibet

    config_file = os.path.join(quodlibet.get_user_dir(), "config")
    quodlibet.init_cli(config_file=config_file)

    try:
        # we want basic commands not to import gtk (doubles process time)
        assert "gi.repository.Gtk" not in sys.modules
        sys.modules["gi.repository.Gtk"] = None
        startup_actions, cmds_todo = process_arguments(argv)
    finally:
        sys.modules.pop("gi.repository.Gtk", None)

    quodlibet.init()

    from quodlibet import app
    from quodlibet.qltk import add_signal_watch

    add_signal_watch(app.quit)

    import quodlibet.player
    import quodlibet.library
    from quodlibet import config
    from quodlibet import browsers
    from quodlibet import util

    app.name = "Quod Libet"
    app.description = _("Music player and music library manager")
    app.id = "io.github.quodlibet.QuodLibet"
    app.process_name = "quodlibet"
    quodlibet.set_application_info(app)

    library_path = os.path.join(quodlibet.get_user_dir(), "songs")

    print_d(f"Initializing main library ({quodlibet.util.path.unexpand(library_path)})")

    library = quodlibet.library.init(library_path)
    app.library = library

    # this assumes that nullbe will always succeed
    from quodlibet.player import PlayerError

    wanted_backend = os.environ.get(
        "QUODLIBET_BACKEND", config.get("player", "backend")
    )

    try:
        player = quodlibet.player.init_player(wanted_backend, app.librarian)
    except PlayerError:
        print_exc()
        player = quodlibet.player.init_player("nullbe", app.librarian)

    app.player = player

    os.environ["PULSE_PROP_media.role"] = "music"
    os.environ["PULSE_PROP_application.icon_name"] = app.icon_name

    browsers.init()

    from quodlibet.qltk.songlist import SongList, get_columns

    headers = get_columns()
    SongList.set_all_column_headers(headers)

    for opt in config.options("header_maps"):
        val = config.get("header_maps", opt)
        util.tags.add(opt, val)

    in_all = (
        "~filename ~uri ~#lastplayed ~#rating ~#playcount ~#skipcount "
        "~#added ~#bitrate ~current ~#laststarted ~basename "
        "~dirname"
    ).split()
    for browser_cls in browsers.browsers:
        if browser_cls.headers is not None:
            browser_cls.headers.extend(in_all)
        browser_cls.init(library)

    pm = quodlibet.init_plugins("no-plugins" in startup_actions)

    if hasattr(player, "init_plugins"):
        player.init_plugins()

    from quodlibet.qltk import unity

    unity.init("io.github.quodlibet.QuodLibet.desktop", player)

    from quodlibet.qltk.songsmenu import SongsMenu

    SongsMenu.init_plugins()

    from quodlibet.util.cover import CoverManager

    app.cover_manager = CoverManager()
    app.cover_manager.init_plugins()

    from quodlibet.plugins.playlist import PLAYLIST_HANDLER

    PLAYLIST_HANDLER.init_plugins()

    from quodlibet.plugins.query import QUERY_HANDLER

    QUERY_HANDLER.init_plugins()

    from gi.repository import GLib

    from quodlibet.commands import registry as cmd_registry, CommandError

    def exec_commands(*args):
        for cmd in cmds_todo:
            try:
                resp = cmd_registry.run(app, *cmd)
            except CommandError:
                pass
            else:
                if resp is not None:
                    print_(resp, end="", flush=True)

    from quodlibet.qltk.quodlibetwindow import QuodLibetWindow, PlayerOptions
    # Call exec_commands after the window is restored, but make sure
    # it's after the mainloop has started so everything is set up.

    app.window = window = QuodLibetWindow(
        library,
        player,
        restore_cb=lambda: GLib.idle_add(exec_commands, priority=GLib.PRIORITY_HIGH),
    )

    app.player_options = PlayerOptions(window)

    from quodlibet.qltk.window import Window

    from quodlibet.plugins.events import EventPluginHandler
    from quodlibet.plugins.gui import UserInterfacePluginHandler

    pm.register_handler(
        EventPluginHandler(library.librarian, player, app.window.songlist)
    )
    pm.register_handler(UserInterfacePluginHandler())

    from quodlibet.mmkeys import MMKeysHandler
    from quodlibet.remote import Remote, RemoteError
    from quodlibet.qltk.tracker import SongTracker, FSInterface

    try:
        from quodlibet.qltk.dbus_ import DBusHandler
    except ImportError:
        DBusHandler = None  # noqa

    mmkeys_handler = MMKeysHandler(app)
    mmkeys_handler.start()

    current_path = os.path.join(quodlibet.get_user_dir(), "current")
    fsiface = FSInterface(current_path, player, library)
    remote = Remote(app, cmd_registry)
    try:
        remote.start()
    except RemoteError:
        exit_(1, True)

    if DBusHandler is not None:
        DBusHandler(player, library)
    tracker = SongTracker(library.librarian, player, window.playlist)

    from quodlibet import session

    session_client = session.init(app)

    quodlibet.enable_periodic_save(save_library=True)

    if "start-playing" in startup_actions or (
        config.getboolean("player", "restore_playing", False)
        and config.getboolean("player", "is_playing", False)
    ):
        player.paused = False

    if "start-hidden" in startup_actions:
        Window.prevent_inital_show(True)

    # restore browser windows
    from quodlibet.qltk.browser import LibraryBrowser

    GLib.idle_add(LibraryBrowser.restore, library, player, priority=GLib.PRIORITY_HIGH)

    def before_quit():
        print_d("Saving active browser state")
        try:
            app.browser.save()
        except NotImplementedError:
            pass

        print_d(f"Shutting down player device {player.version_info!r}.")
        player.destroy()

    quodlibet.run(window, before_quit=before_quit)

    app.player_options.destroy()
    quodlibet.finish_first_session("quodlibet")
    mmkeys_handler.quit()
    remote.stop()
    fsiface.destroy()

    tracker.destroy()
    quodlibet.library.save()
    quodlibet.library.destroy()
    config.save()

    session_client.close()

    print_d("Finished shutdown.")

    if app.is_restarting:
        os.execv(sys.executable, [sys.executable] + sys.argv)



# Copyright 2010,2012 Christoph Reiter <reiter.christoph@gmail.com>
#           2022-24 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys

if os.name == "nt" or sys.platform == "darwin":
    from quodlibet.plugins import PluginNotSupportedError

    raise PluginNotSupportedError

from gi.repository import Gtk

import dbus

try:
    import indicate
except ImportError:
    indicate = None

from quodlibet import _
from quodlibet import app
from quodlibet import config
from quodlibet import qltk
from quodlibet.qltk.ccb import ConfigSwitch
from quodlibet.qltk import Icons
from quodlibet.plugins.events import EventPlugin

from .mpris2 import MPRIS2


class MPRIS(EventPlugin):
    PLUGIN_ID = "mpris"
    PLUGIN_NAME = _("Linux Desktop Integration (MPRIS D-Bus)")
    PLUGIN_DESC_MARKUP = _(
        "⏯️ Allows control of Quod Libet using the "
        '<a href="https://mpris2.readthedocs.io/en/latest/">MPRIS 2</a> '
        "D-Bus Interface.\n"
        "This allows various Linux desktop integrations (e.g. multimedia keys)."
    )
    PLUGIN_ICON = Icons.MEDIA_PLAYBACK_START

    def PluginPreferences(self, parent):
        box = Gtk.HBox()
        ccb = ConfigSwitch(
            _("Hide main window on close"),
            "plugins",
            "mpris_window_hide",
            populate=True,
        )
        box.pack_start(qltk.Frame(_("Preferences"), child=ccb), True, True, 0)
        return box

    def __do_hide(self):
        return config.getboolean("plugins", "mpris_window_hide", False)

    def __window_delete(self, win, event):
        if self.__do_hide():
            win.hide()
            return True
        return None

    def enabled(self):
        self.__sig = app.window.connect("delete-event", self.__window_delete)

        self.objects = []
        for service in [MPRIS2]:
            try:
                self.objects.append(service())
            except dbus.DBusException:
                pass

        # Needed for sound menu support in some older Ubuntu versions
        if indicate:
            self.__indicate_server = s = indicate.indicate_server_ref_default()
            s.set_type("music.quodlibet")
            s.set_desktop_file(
                "/usr/share/applications/io.github.quodlibet.QuodLibet.desktop"
            )
            s.show()

    def disabled(self):
        if indicate:
            self.__indicate_server.hide()

        for obj in self.objects:
            obj.remove_from_connection()
        self.objects = []

        import gc

        gc.collect()

        app.window.disconnect(self.__sig)

    def plugin_on_paused(self):
        for obj in self.objects:
            obj.paused()

    def plugin_on_unpaused(self):
        for obj in self.objects:
            obj.unpaused()

    def plugin_on_song_started(self, song):
        for obj in self.objects:
            obj.song_started(song)

    def plugin_on_song_ended(self, song, skipped):
        for obj in self.objects:
            obj.song_ended(song, skipped)
